# Copyright 2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test class for iRMC eLCM functionality.
"""

from unittest import mock

from oslo_utils import encodeutils
import requests
from requests_mock.contrib import fixture as rm_fixture
import testtools

from scciclient.irmc import elcm
from scciclient.irmc import scci


class ELCMTestCase(testtools.TestCase):
    """Tests for eLCM"""

    RESPONSE_TEMPLATE = ('Date: Mon, 07 Dec 2015 17:10:55 GMT\n'
                         'Server: iRMC S4 Webserver\n'
                         'Content-Length: 123\n'
                         'Content-Type: application/json; charset=UTF-8\n'
                         '\r\n'
                         '{\n'
                         '%(json_text)s\n'
                         '}')

    def setUp(self):
        super(ELCMTestCase, self).setUp()

        self.requests_mock = self.useFixture(rm_fixture.Fixture())

        self.irmc_info = {
            'irmc_address': '10.124.196.159',
            'irmc_username': 'admin',
            'irmc_password': 'admin0',
            'irmc_port': 80,
            'irmc_auth_method': 'basic',
            'irmc_client_timeout': 60,
        }

        self.raid_info = {
            'Server': {
                'HWConfigurationIrmc': {
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': None,
                                'LogicalDrives': None
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }

    def _create_server_url(self, path, port=None):
        scheme = 'unknown'

        if port is None:
            port = self.irmc_info['irmc_port']
        if port == 80:
            scheme = 'http'
        elif port == 443:
            scheme = 'https'

        return ('%(scheme)s://%(addr)s%(path)s' %
                {'scheme': scheme,
                 'addr': self.irmc_info['irmc_address'],
                 'path': path})

    def _create_server_response(self, content):
        response = requests.Response()
        response._content = encodeutils.safe_encode(content)
        response.encoding = 'utf-8'
        return response

    def test__parse_elcm_response_body_as_json_empty(self):
        response = self._create_server_response('')
        self.assertRaises(elcm.ELCMInvalidResponse,
                          elcm._parse_elcm_response_body_as_json,
                          response=response)

    def test__parse_elcm_response_body_as_json_invalid(self):
        content = 'abc123'
        response = self._create_server_response(content)
        self.assertRaises(elcm.ELCMInvalidResponse,
                          elcm._parse_elcm_response_body_as_json,
                          response=response)

    def test__parse_elcm_response_body_as_json_missing_empty_line(self):
        content = ('key1:val1\nkey2:val2\n'
                   '{"1":1,"2":[123, "abc"],"3":3}')
        response = self._create_server_response(content)
        self.assertRaises(elcm.ELCMInvalidResponse,
                          elcm._parse_elcm_response_body_as_json,
                          response=response)

    def test__parse_elcm_response_body_as_json_ok(self):
        content = ('key1:val1\nkey2:val2\n'
                   '\r\n'
                   '{"1":1,"2":[123, "abc"],"3":3}')
        response = self._create_server_response(content)
        result = elcm._parse_elcm_response_body_as_json(response)

        expected = {
            "1": 1,
            "2": [123, "abc"],
            "3": 3
        }
        self.assertEqual(expected, result)

    def test_elcm_request_protocol_http_ok(self):
        irmc_info = dict(self.irmc_info)
        irmc_info['irmc_port'] = 80

        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT,
                                    port=80),
            text=self.RESPONSE_TEMPLATE % {'json_text': 'abc123'})

        response = elcm.elcm_request(
            irmc_info,
            method='GET',
            path=elcm.URL_PATH_PROFILE_MGMT)

        self.assertEqual(200, response.status_code)

        expected = self.RESPONSE_TEMPLATE % {'json_text': 'abc123'}
        self.assertEqual(expected, response.text)

    def test_elcm_request_protocol_https_ok(self):
        irmc_info = dict(self.irmc_info)
        irmc_info['irmc_port'] = 443

        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT,
                                    port=443),
            text=self.RESPONSE_TEMPLATE % {'json_text': 'abc123'})

        response = elcm.elcm_request(
            irmc_info,
            method='GET',
            path=elcm.URL_PATH_PROFILE_MGMT)

        self.assertEqual(200, response.status_code)

        expected = self.RESPONSE_TEMPLATE % {'json_text': 'abc123'}
        self.assertEqual(expected, response.text)

    def test_elcm_request_unsupported_port(self):
        irmc_info = dict(self.irmc_info)
        irmc_info['irmc_port'] = 22

        e = self.assertRaises(scci.SCCIInvalidInputError,
                              elcm.elcm_request,
                              irmc_info,
                              method='GET',
                              path=elcm.URL_PATH_PROFILE_MGMT)

        auth_method = self.irmc_info['irmc_auth_method']
        self.assertEqual((("Invalid port %(port)d or "
                           "auth_method for method %(auth_method)s") %
                          {'port': 22,
                           'auth_method': auth_method}), str(e))

    def test_elcm_request_auth_method_basic(self):
        irmc_info = dict(self.irmc_info)
        irmc_info['irmc_auth_method'] = 'basic'

        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT),
            status_code=200,
            text='ok')

        response = elcm.elcm_request(
            irmc_info,
            method='GET',
            path=elcm.URL_PATH_PROFILE_MGMT)

        self.assertEqual(200, response.status_code)
        self.assertEqual('ok', response.text)

    def test_elcm_request_auth_method_digest(self):
        irmc_info = dict(self.irmc_info)
        irmc_info['irmc_auth_method'] = 'digest'

        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT),
            status_code=200,
            text='ok')

        response = elcm.elcm_request(
            irmc_info,
            method='GET',
            path=elcm.URL_PATH_PROFILE_MGMT)

        self.assertEqual(200, response.status_code)
        self.assertEqual('ok', response.text)

    def test_elcm_request_unsupported_auth_method(self):
        irmc_info = dict(self.irmc_info)
        irmc_info['irmc_auth_method'] = 'unknown'

        e = self.assertRaises(scci.SCCIInvalidInputError,
                              elcm.elcm_request,
                              irmc_info,
                              method='GET',
                              path=elcm.URL_PATH_PROFILE_MGMT)

        port = self.irmc_info['irmc_port']
        self.assertEqual((("Invalid port %(port)d or "
                           "auth_method for method %(auth_method)s") %
                          {'port': port,
                           'auth_method': 'unknown'}), str(e))

    def test_elcm_request_auth_failed(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT),
            status_code=401,
            text='401 Unauthorized')

        e = self.assertRaises(scci.SCCIClientError,
                              elcm.elcm_request,
                              self.irmc_info,
                              method='GET',
                              path=elcm.URL_PATH_PROFILE_MGMT)

        self.assertEqual('UNAUTHORIZED', str(e))

    def test_elcm_profile_get_versions_failed(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT) + 'version',
            status_code=503)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_profile_get_versions,
                          self.irmc_info)

    def test_elcm_profile_get_versions_ok(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT) + 'version',
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Server":{'
                    '  "@Version": "1.01",'
                    '  "AdapterConfigIrmc": {'
                    '    "@Version": "1.00"'
                    '   },'
                    '  "SystemConfig": {'
                    '    "BiosConfig": {'
                    '      "@Version": "1.02"'
                    '    }'
                    '  }'
                    '}'}))

        result = elcm.elcm_profile_get_versions(self.irmc_info)

        expected = {
            "Server": {
                "@Version": "1.01",
                "AdapterConfigIrmc": {
                    "@Version": "1.00"
                },
                "SystemConfig": {
                    "BiosConfig": {
                        "@Version": "1.02"
                    }
                }
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_profile_get_not_found(self):
        profile_name = elcm.PROFILE_BIOS_CONFIG
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url((elcm.URL_PATH_PROFILE_MGMT +
                                     profile_name)),
            status_code=404)

        self.assertRaises(elcm.ELCMProfileNotFound,
                          elcm.elcm_profile_get,
                          self.irmc_info,
                          profile_name=profile_name)

    def test_elcm_profile_list_failed(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT),
            status_code=503)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_profile_list,
                          self.irmc_info)

    def test_elcm_profile_list_ok(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT),
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Links":{'
                    '  "profileStore":['
                    '    {"@odata.id": "id1"},'
                    '    {"@odata.id": "id2"}'
                    '    ]'
                    '}'}))

        result = elcm.elcm_profile_list(self.irmc_info)

        expected = {
            "Links": {
                "profileStore": [
                    {"@odata.id": "id1"},
                    {"@odata.id": "id2"}
                ]
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_profile_get_failed(self):
        profile_name = elcm.PROFILE_BIOS_CONFIG
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url((elcm.URL_PATH_PROFILE_MGMT +
                                     profile_name)),
            status_code=500)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_profile_get,
                          self.irmc_info,
                          profile_name=profile_name)

    def test_elcm_profile_get_ok(self):
        profile_name = elcm.PROFILE_BIOS_CONFIG
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url((elcm.URL_PATH_PROFILE_MGMT +
                                     profile_name)),
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Server":{'
                    '  "SystemConfig":{'
                    '    "BiosConfig":{'
                    '       "key":"val"'
                    '    }'
                    '  }'
                    '}'}))

        result = elcm.elcm_profile_get(
            self.irmc_info,
            profile_name=profile_name)

        expected = {
            "Server": {
                "SystemConfig": {
                    "BiosConfig": {
                        "key": "val"
                    }
                }
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_profile_create_failed(self):
        param_path = elcm.PARAM_PATH_BIOS_CONFIG
        self.requests_mock.register_uri(
            'POST',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT + 'get'),
            status_code=200)  # Success code is 202

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_profile_create,
                          self.irmc_info,
                          param_path=param_path)

    def test_elcm_profile_create_ok(self):
        param_path = elcm.PARAM_PATH_BIOS_CONFIG
        self.requests_mock.register_uri(
            'POST',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT + 'get'),
            status_code=202,  # Success code is 202
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Session":{'
                    '  "Id": 123,'
                    '  "Status": "activated"'
                    '}'}))

        result = elcm.elcm_profile_create(self.irmc_info,
                                          param_path=param_path)

        expected = {
            "Session": {
                "Id": 123,
                "Status": "activated",
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_profile_set_failed(self):
        input_data = {
            "Server": {
                "SystemConfig": {
                    "BiosConfig": {
                        "key": "val"
                    }
                }
            }
        }
        self.requests_mock.register_uri(
            'POST',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT + 'set'),
            status_code=200)  # Success code is 202

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_profile_set,
                          self.irmc_info,
                          input_data=input_data)

    def test_elcm_profile_set_ok(self):
        input_data = {
            "Server": {
                "SystemConfig": {
                    "BiosConfig": {
                        "key": "val"
                    }
                }
            }
        }
        self.requests_mock.register_uri(
            'POST',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT + 'set'),
            status_code=202,  # Success code is 202
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Session":{'
                    '  "Id": 123,'
                    '  "Status": "activated"'
                    '}'}))

        result = elcm.elcm_profile_set(self.irmc_info,
                                       input_data=input_data)

        expected = {
            "Session": {
                "Id": 123,
                "Status": "activated",
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_profile_set_with_raid_config_failed(self):
        input_data = {
            "Server": {
                "HWConfigurationIrmc": {
                    "Adapters": {
                        "RAIDAdapter": [
                            {
                                "key": "val"
                            }
                        ]
                    }
                }
            }
        }
        self.requests_mock.register_uri(
            'POST',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT + 'set'),
            status_code=400)  # Success code is 202

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_profile_set,
                          self.irmc_info,
                          input_data=input_data)

    def test_elcm_profile_set_with_raid_config_ok(self):
        input_data = {
            "Server": {
                "HWConfigurationIrmc": {
                    "Adapters": {
                        "RAIDAdapter": [
                            {
                                "key": "val"
                            }
                        ]
                    }
                }
            }
        }
        self.requests_mock.register_uri(
            'POST',
            self._create_server_url(elcm.URL_PATH_PROFILE_MGMT + 'set'),
            status_code=202,  # Success code is 202
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Session":{'
                    '  "Id": 123,'
                    '  "Status": "activated"'
                    '}'}))

        result = elcm.elcm_profile_set(self.irmc_info,
                                       input_data=input_data)

        expected = {
            "Session": {
                "Id": 123,
                "Status": "activated",
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_profile_delete_not_found(self):
        profile_name = elcm.PROFILE_BIOS_CONFIG
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url((elcm.URL_PATH_PROFILE_MGMT +
                                     profile_name)),
            status_code=404)

        self.assertRaises(elcm.ELCMProfileNotFound,
                          elcm.elcm_profile_delete,
                          self.irmc_info,
                          profile_name=profile_name)

    def test_elcm_profile_delete_failed(self):
        profile_name = elcm.PROFILE_BIOS_CONFIG
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url((elcm.URL_PATH_PROFILE_MGMT +
                                     profile_name)),
            status_code=500)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_profile_delete,
                          self.irmc_info,
                          profile_name=profile_name)

    def test_elcm_profile_delete_ok(self):
        profile_name = elcm.PROFILE_BIOS_CONFIG
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url((elcm.URL_PATH_PROFILE_MGMT +
                                     profile_name)),
            text='ok')

        result = elcm.elcm_profile_delete(self.irmc_info,
                                          profile_name=profile_name)

        self.assertIsNone(result)

    def test_elcm_session_list_failed(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url('/sessionInformation/'),
            status_code=500)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_session_list,
                          self.irmc_info)

    def test_elcm_session_list_empty(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url('/sessionInformation/'),
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"SessionList":{'
                    '  "Contains":['
                    '  ]'
                    '}'}))

        result = elcm.elcm_session_list(self.irmc_info)

        expected = {
            "SessionList": {
                "Contains": [
                ]
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_session_list_ok(self):
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url('/sessionInformation/'),
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"SessionList":{'
                    '  "Contains":['
                    '    { "Id": 1, "Name": "name1" },'
                    '    { "Id": 2, "Name": "name2" },'
                    '    { "Id": 3, "Name": "name3" }'
                    '  ]'
                    '}'}))

        result = elcm.elcm_session_list(self.irmc_info)

        expected = {
            "SessionList": {
                "Contains": [
                    {"Id": 1, "Name": "name1"},
                    {"Id": 2, "Name": "name2"},
                    {"Id": 3, "Name": "name3"}
                ]
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_session_get_status_not_found(self):
        session_id = 123
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(('/sessionInformation/%s/status' %
                                     session_id)),
            status_code=404)

        self.assertRaises(elcm.ELCMSessionNotFound,
                          elcm.elcm_session_get_status,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_get_status_failed(self):
        session_id = 123
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(('/sessionInformation/%s/status' %
                                     session_id)),
            status_code=500)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_session_get_status,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_get_status_ok(self):
        session_id = 123
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(('/sessionInformation/%s/status' %
                                     session_id)),
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Session":{'
                    '  "Id": 123,'
                    '  "Status": "abc123"'
                    '}'}))

        result = elcm.elcm_session_get_status(self.irmc_info,
                                              session_id=session_id)

        expected = {
            "Session": {
                "Id": 123,
                "Status": "abc123",
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_session_get_log_not_found(self):
        session_id = 123
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(('/sessionInformation/%s/log' %
                                     session_id)),
            status_code=404)

        self.assertRaises(elcm.ELCMSessionNotFound,
                          elcm.elcm_session_get_log,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_get_log_failed(self):
        session_id = 123
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(('/sessionInformation/%s/log' %
                                     session_id)),
            status_code=500)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_session_get_log,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_get_log_ok(self):
        session_id = 123
        self.requests_mock.register_uri(
            'GET',
            self._create_server_url(('/sessionInformation/%s/log' %
                                     session_id)),
            text=(self.RESPONSE_TEMPLATE % {
                'json_text':
                    '"Session":{'
                    '  "Id": 123,'
                    '  "A_Param": "abc123"'
                    '}'}))

        result = elcm.elcm_session_get_log(self.irmc_info,
                                           session_id=session_id)

        expected = {
            "Session": {
                "Id": 123,
                "A_Param": "abc123",
            }
        }
        self.assertEqual(expected, result)

    def test_elcm_session_terminate_not_found(self):
        session_id = 123
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url(('/sessionInformation/%s/terminate' %
                                     session_id)),
            status_code=404)

        self.assertRaises(elcm.ELCMSessionNotFound,
                          elcm.elcm_session_terminate,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_terminate_failed(self):
        session_id = 123
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url(('/sessionInformation/%s/terminate' %
                                     session_id)),
            status_code=500)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_session_terminate,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_terminate_ok(self):
        session_id = 123
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url(('/sessionInformation/%s/terminate' %
                                     session_id)),
            text='ok')

        result = elcm.elcm_session_terminate(self.irmc_info,
                                             session_id=session_id)

        self.assertIsNone(result)

    def test_elcm_session_delete_not_found(self):
        session_id = 123
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url(('/sessionInformation/%s/remove' %
                                     session_id)),
            status_code=404)

        self.assertRaises(elcm.ELCMSessionNotFound,
                          elcm.elcm_session_delete,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_delete_failed(self):
        session_id = 123
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url(('/sessionInformation/%s/remove' %
                                     session_id)),
            status_code=500)

        self.assertRaises(scci.SCCIClientError,
                          elcm.elcm_session_delete,
                          self.irmc_info,
                          session_id=session_id)

    def test_elcm_session_delete_ok(self):
        session_id = 123
        self.requests_mock.register_uri(
            'DELETE',
            self._create_server_url(('/sessionInformation/%s/remove' %
                                     session_id)),
            text='ok')

        result = elcm.elcm_session_delete(self.irmc_info,
                                          session_id=session_id)

        self.assertIsNone(result)

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_get')
    @mock.patch.object(elcm, 'elcm_session_delete')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    @mock.patch.object(elcm.time, 'sleep')
    def test__process_session_data_get_ok(self, mock_sleep,
                                          mock_session_get,
                                          mock_session_delete,
                                          mock_profile_get,
                                          mock_profile_delete):
        session_id = 123
        expected_bios_cfg = {
            'Server': {
                'SystemConfig': {
                    'BiosConfig': {
                        'key1': 'val1'
                    }
                }
            }
        }
        mock_session_get.side_effect = [
            {'Session': {'Id': session_id,
                         'Status': 'activated'}},
            {'Session': {'Id': session_id,
                         'Status': 'running'}},
            {'Session': {'Id': session_id,
                         'Status': 'terminated regularly'}}]
        mock_profile_get.return_value = expected_bios_cfg

        result = elcm._process_session_data(irmc_info=self.irmc_info,
                                            operation='BACKUP_BIOS',
                                            session_id=session_id)
        self.assertEqual(expected_bios_cfg, result['bios_config'])

        mock_session_get.assert_has_calls([
            mock.call(irmc_info=self.irmc_info, session_id=session_id),
            mock.call(irmc_info=self.irmc_info, session_id=session_id),
            mock.call(irmc_info=self.irmc_info, session_id=session_id)])
        mock_profile_get.assert_called_once_with(
            irmc_info=self.irmc_info,
            profile_name=elcm.PROFILE_BIOS_CONFIG)

        self.assertEqual(2, mock_sleep.call_count)
        self.assertEqual(1, mock_session_delete.call_count)
        self.assertEqual(0, mock_profile_delete.call_count)

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_get')
    @mock.patch.object(elcm, 'elcm_session_delete')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    @mock.patch.object(elcm.time, 'sleep')
    def test__process_session_data_set_ok(self, mock_sleep,
                                          mock_session_get,
                                          mock_session_delete,
                                          mock_profile_get,
                                          mock_profile_delete):
        session_id = 123
        mock_session_get.side_effect = [
            {'Session': {'Id': session_id,
                         'Status': 'activated'}},
            {'Session': {'Id': session_id,
                         'Status': 'running'}},
            {'Session': {'Id': session_id,
                         'Status': 'terminated regularly'}}]

        elcm._process_session_data(irmc_info=self.irmc_info,
                                   operation='RESTORE_BIOS',
                                   session_id=session_id)

        mock_session_get.assert_has_calls([
            mock.call(irmc_info=self.irmc_info, session_id=session_id),
            mock.call(irmc_info=self.irmc_info, session_id=session_id),
            mock.call(irmc_info=self.irmc_info, session_id=session_id)])
        mock_profile_get.assert_not_called()
        self.assertEqual(2, mock_sleep.call_count)
        self.assertEqual(1, mock_session_delete.call_count)
        self.assertEqual(0, mock_profile_delete.call_count)

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_get')
    @mock.patch.object(elcm, 'elcm_session_delete')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    @mock.patch.object(elcm.time, 'sleep')
    def test__process_session_data_timeout(self, mock_sleep,
                                           mock_session_get,
                                           mock_session_delete,
                                           mock_profile_get,
                                           mock_profile_delete):
        session_id = 123
        mock_session_get.return_value = {'Session': {'Id': session_id,
                                                     'Status': 'running'}}

        self.assertRaises(elcm.ELCMSessionTimeout,
                          elcm._process_session_data,
                          irmc_info=self.irmc_info,
                          operation='BACKUP_BIOS',
                          session_id=session_id,
                          session_timeout=0.5)

        self.assertEqual(True, mock_sleep.called)
        self.assertEqual(True, mock_session_get.called)
        mock_profile_get.assert_not_called()
        mock_session_delete.assert_not_called()
        mock_profile_delete.assert_not_called()

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_session_delete')
    @mock.patch.object(elcm, 'elcm_session_get_log')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    def test__process_session_data_error(self,
                                         mock_session_get,
                                         mock_session_get_log,
                                         mock_session_delete,
                                         mock_profile_delete):
        session_id = 123
        mock_session_get.return_value = {'Session': {'Id': session_id,
                                                     'Status': 'error'}}

        self.assertRaises(scci.SCCIClientError,
                          elcm._process_session_data,
                          irmc_info=self.irmc_info,
                          operation='RESTORE_BIOS',
                          session_id=session_id,
                          session_timeout=0.5)

        self.assertEqual(True, mock_session_get.called)
        self.assertEqual(True, mock_session_get_log.called)
        mock_session_delete.assert_not_called()
        mock_profile_delete.assert_not_called()

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_create')
    @mock.patch.object(elcm, 'elcm_profile_get')
    @mock.patch.object(elcm, 'elcm_session_delete')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    @mock.patch.object(elcm.time, 'sleep')
    def test_backup_bios_config_ok(self, mock_sleep, mock_session_get,
                                   mock_session_delete, mock_profile_get,
                                   mock_profile_create, mock_profile_delete):
        session_id = 123
        expected_bios_cfg = {
            'Server': {
                'SystemConfig': {
                    'BiosConfig': {
                        'key1': 'val1'
                    }
                }
            }
        }
        mock_session_get.side_effect = [
            {'Session': {'Id': session_id,
                         'Status': 'activated'}},
            {'Session': {'Id': session_id,
                         'Status': 'running'}},
            {'Session': {'Id': session_id,
                         'Status': 'terminated regularly'}}]
        mock_profile_get.return_value = expected_bios_cfg
        mock_profile_create.return_value = {'Session': {'Id': session_id,
                                                        'Status': 'activated'}}

        result = elcm.backup_bios_config(irmc_info=self.irmc_info)
        self.assertEqual(expected_bios_cfg, result['bios_config'])

        self.assertEqual(0, mock_sleep.call_count)
        self.assertEqual(False, mock_session_get.called)
        self.assertEqual(0, mock_session_delete.call_count)
        self.assertEqual(1, mock_profile_get.call_count)
        self.assertEqual(0, mock_profile_create.call_count)
        self.assertEqual(0, mock_profile_delete.call_count)

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_get')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, 'elcm_session_delete')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    @mock.patch.object(elcm.time, 'sleep')
    def _test_restore_bios_config_ok(self, mock_sleep, mock_session_get,
                                     mock_session_delete, mock_profile_set,
                                     mock_profile_get, mock_profile_delete,
                                     bios_cfg):
        session_id = 123
        mock_session_get.side_effect = [
            {'Session': {'Id': session_id,
                         'Status': 'activated'}},
            {'Session': {'Id': session_id,
                         'Status': 'running'}},
            {'Session': {'Id': session_id,
                         'Status': 'terminated regularly'}}]
        mock_profile_set.return_value = {'Session': {'Id': session_id,
                                                     'Status': 'activated'}}

        elcm.restore_bios_config(irmc_info=self.irmc_info,
                                 bios_config=bios_cfg)

        self.assertEqual(2, mock_sleep.call_count)
        self.assertEqual(True, mock_session_get.called)
        self.assertEqual(1, mock_session_delete.call_count)
        self.assertEqual(0, mock_profile_get.call_count)
        self.assertEqual(1, mock_profile_set.call_count)
        self.assertEqual(0, mock_profile_delete.call_count)

    def test_restore_bios_config_ok_with_dict(self):
        bios_cfg = {
            'Server': {
                'SystemConfig': {
                    'BiosConfig': {
                        'key1': 'val1'
                    }
                }
            }
        }
        self._test_restore_bios_config_ok(bios_cfg=bios_cfg)

    def test_restore_bios_config_ok_with_str(self):
        bios_cfg = ('{"Server":'
                    '  {"SystemConfig":'
                    '    {"BiosConfig":'
                    '      {'
                    '        "key1": "val1"'
                    '      }'
                    '    }'
                    '  }'
                    '}')
        self._test_restore_bios_config_ok(bios_cfg=bios_cfg)

    def _test_restore_bios_config_invalid_input(self, bios_cfg):
        self.assertRaises(scci.SCCIInvalidInputError,
                          elcm.restore_bios_config,
                          irmc_info=self.irmc_info,
                          bios_config=bios_cfg)

    def test_restore_bios_config_invalid_input_dict(self):
        bios_cfg = {
            'Server': {
                'SystemConfig': {
                }
            }
        }
        self._test_restore_bios_config_invalid_input(bios_cfg=bios_cfg)

    def test_restore_bios_config_invalid_input_str(self):
        bios_cfg = '{"key": "val"}'
        self._test_restore_bios_config_invalid_input(bios_cfg=bios_cfg)

    @mock.patch.object(elcm, 'backup_bios_config')
    def test_get_secure_boot_mode_true(self, backup_bios_config_mock):
        backup_bios_config_mock.return_value = {
            'bios_config': {
                'Server': {
                    'SystemConfig': {
                        'BiosConfig': {
                            'SecurityConfig': {
                                'SecureBootControlEnabled': True
                            }
                        }
                    }
                }
            }
        }
        result = elcm.get_secure_boot_mode(irmc_info=self.irmc_info)
        self.assertEqual(True, result)
        backup_bios_config_mock.assert_called_once_with(
            irmc_info=self.irmc_info)

    @mock.patch.object(elcm, 'backup_bios_config')
    def test_get_secure_boot_mode_false(self, backup_bios_config_mock):
        backup_bios_config_mock.return_value = {
            'bios_config': {
                'Server': {
                    'SystemConfig': {
                        'BiosConfig': {
                            'SecurityConfig': {
                                'SecureBootControlEnabled': False
                            }
                        }
                    }
                }
            }
        }
        result = elcm.get_secure_boot_mode(irmc_info=self.irmc_info)
        self.assertEqual(False, result)
        backup_bios_config_mock.assert_called_once_with(
            irmc_info=self.irmc_info)

    @mock.patch.object(elcm, 'backup_bios_config')
    def test_get_secure_boot_mode_fail(self, backup_bios_config_mock):
        backup_bios_config_mock.return_value = {
            'bios_config': {
                'Server': {
                    'SystemConfig': {
                        'BiosConfig': {
                            'SecurityConfig': {
                                'FlashWriteEnabled': False
                            }
                        }
                    }
                }
            }
        }

        self.assertRaises(elcm.SecureBootConfigNotFound,
                          elcm.get_secure_boot_mode,
                          irmc_info=self.irmc_info)
        backup_bios_config_mock.assert_called_once_with(
            irmc_info=self.irmc_info)

    @mock.patch.object(elcm, 'restore_bios_config')
    def test_set_secure_boot_mode_true(self, restore_bios_config_mock):
        elcm.set_secure_boot_mode(irmc_info=self.irmc_info, enable=True)
        bios_config_data = {
            'Server': {
                '@Version': '1.01',
                'SystemConfig': {
                    'BiosConfig': {
                        '@Version': '1.01',
                        'SecurityConfig': {
                            'SecureBootControlEnabled': True
                        }
                    }
                }
            }
        }
        restore_bios_config_mock.assert_called_once_with(
            irmc_info=self.irmc_info, bios_config=bios_config_data)

    @mock.patch.object(elcm, 'restore_bios_config')
    def test_set_secure_boot_mode_false(self, restore_bios_config_mock):
        elcm.set_secure_boot_mode(irmc_info=self.irmc_info, enable=False)
        bios_config_data = {
            'Server': {
                '@Version': '1.01',
                'SystemConfig': {
                    'BiosConfig': {
                        '@Version': '1.01',
                        'SecurityConfig': {
                            'SecureBootControlEnabled': False
                        }
                    }
                }
            }
        }
        restore_bios_config_mock.assert_called_once_with(
            irmc_info=self.irmc_info, bios_config=bios_config_data)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_without_logical_drives(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):
        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        target_raid_config = {
            'logical_disks': [
                {
                    'size_gb': 100,
                    'raid_level': '1',
                }
            ]
        }
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': []
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '1',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        # In case of "LogicalDrives" doesn't exist
        del self.raid_info['Server']['HWConfigurationIrmc'][
            'Adapters']['RAIDAdapter'][0]['LogicalDrives']
        raid_info_mock.return_value = self.raid_info

        elcm.create_raid_configuration(
            self.irmc_info, target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        raid_info_mock.assert_called_once_with(self.irmc_info)
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_without_arrays_info_and_physical_disks(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': []
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '1',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 100,
                        'raid_level': '1',
                    }
                ]
        }

        elcm.create_raid_configuration(
            self.irmc_info, target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        raid_info_mock.assert_called_once_with(self.irmc_info)
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_with_arrays_info_and_without_physical_disks(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': []
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '0',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                        },
                                        {
                                            '@Number': 1,
                                            '@Action': 'Create',
                                            'RaidLevel': '1',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 100,
                        'raid_level': '0',
                    },
                    {
                        'size_gb': 100,
                        'raid_level': '1',
                    }
                ]
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        elcm.create_raid_configuration(
            self.irmc_info, target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)
        raid_info_mock.assert_called_once_with(self.irmc_info)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_with_physical_disks_and_without_array_info(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': [
                                        {
                                            '@Number': 0,
                                            '@ConfigurationType': 'Setting',
                                            'PhysicalDiskRefs': {
                                                'PhysicalDiskRef': [
                                                    {
                                                        '@Number': '0'
                                                    },
                                                    {
                                                        '@Number': '1'
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '1',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                            'ArrayRefs': {
                                                'ArrayRef': [
                                                    {
                                                        '@Number': 0
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        target_raid_config = {
            "logical_disks":
                [
                    {
                        'size_gb': 100,
                        "raid_level": "1",
                        "physical_disks": [
                            "0",
                            "1"
                        ],
                    },
                ]
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        elcm.create_raid_configuration(
            self.irmc_info, target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)
        raid_info_mock.assert_called_once_with(self.irmc_info)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_along_with_physical_disks_and_array_info(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': [
                                        {
                                            '@Number': 0,
                                            '@ConfigurationType': 'Setting',
                                            'PhysicalDiskRefs': {
                                                'PhysicalDiskRef': [
                                                    {
                                                        '@Number': '0'
                                                    },
                                                    {
                                                        '@Number': '1'
                                                    }
                                                ]
                                            }
                                        },
                                        {
                                            '@Number': 1,
                                            '@ConfigurationType': 'Setting',
                                            'PhysicalDiskRefs': {
                                                'PhysicalDiskRef': [
                                                    {
                                                        '@Number': '4'
                                                    },
                                                    {
                                                        '@Number': '5'
                                                    }
                                                ]
                                            }
                                        },
                                    ]
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '0',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                            'ArrayRefs': {
                                                'ArrayRef': [
                                                    {
                                                        '@Number': 0
                                                    }
                                                ]
                                            }
                                        },
                                        {
                                            '@Number': 1,
                                            '@Action': 'Create',
                                            'RaidLevel': '1',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                            'ArrayRefs': {
                                                'ArrayRef': [
                                                    {
                                                        '@Number': 1
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 100,
                        'raid_level': '0',
                        'physical_disks': [
                            '0',
                            '1'
                        ]
                    },
                    {
                        'size_gb': 100,
                        'raid_level': '1',
                        'physical_disks': [
                            '4',
                            '5'
                        ]
                    },
                ]
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        elcm.create_raid_configuration(
            self.irmc_info, target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)
        raid_info_mock.assert_called_once_with(self.irmc_info)

    @mock.patch.object(elcm, 'delete_raid_configuration')
    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_with_exist_raid_config(
            self, session_mock, elcm_profile_set_mock, raid_info_mock,
            delete_raid_mock):

        raid_info_mock.return_value = {
            'Server': {
                'HWConfigurationIrmc': {
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': None,
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': None,
                                            'RaidLevel': '1'
                                        },
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        expected_raid_call = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': []
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '1',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            }
                                        },
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }

        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 100,
                        'raid_level': '1'
                    },
                ]
        }

        elcm.create_raid_configuration(self.irmc_info, target_raid_config)

        elcm_profile_set_mock.assert_called_once_with(self.irmc_info,
                                                      expected_raid_call)
        delete_raid_mock.assert_called_once_with(self.irmc_info)
        session_mock.assert_has_calls([mock.call(self.irmc_info, operation,
                                      session_id, session_timeout)])
        raid_info_mock.assert_has_calls([mock.call(self.irmc_info)])

    def test_create_raid_config_without_logical_disk(self):

        target_raid_config = {
            'logical_disks': []
        }

        self.assertRaises(elcm.ELCMValueError,
                          elcm.create_raid_configuration,
                          irmc_info=self.irmc_info,
                          target_raid_config=target_raid_config)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_with_raid_level_is_max(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': []
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '0',
                                            'InitMode': 'fast'
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 'MAX',
                        'raid_level': '0'
                    },
                ]
        }

        elcm.create_raid_configuration(self.irmc_info, target_raid_config)
        elcm_profile_set_mock.assert_called_once_with(self.irmc_info,
                                                      expected_input)
        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_config_hybrid_in_target_raid_config(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': [
                                        {
                                            '@Number': 1,
                                            '@ConfigurationType': 'Setting',
                                            'PhysicalDiskRefs': {
                                                'PhysicalDiskRef': [
                                                    {
                                                        '@Number': '0'
                                                    },
                                                    {
                                                        '@Number': '1'
                                                    }
                                                ]
                                            }
                                        },
                                    ]
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '0',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },

                                        },
                                        {
                                            '@Number': 1,
                                            '@Action': 'Create',
                                            'RaidLevel': '1',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                            'ArrayRefs': {
                                                'ArrayRef': [
                                                    {
                                                        '@Number': 1
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 100,
                        'raid_level': '0',
                    },
                    {
                        'size_gb': 100,
                        'raid_level': '1',
                        'physical_disks': [
                            '0',
                            '1'
                        ]
                    },
                ]
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        elcm.create_raid_configuration(
            self.irmc_info, target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)
        raid_info_mock.assert_called_once_with(self.irmc_info)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_10_in_target_raid_config(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': []
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '10',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            }
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }

        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 100,
                        'raid_level': '10',
                    }
                ]
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        elcm.create_raid_configuration(self.irmc_info,
                                       target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)
        raid_info_mock.assert_called_once_with(self.irmc_info)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_create_raid_50_in_target_raid_config(
            self, session_mock, elcm_profile_set_mock, raid_info_mock):

        raid_info_mock.return_value = self.raid_info
        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'Arrays': {
                                    'Array': []
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Create',
                                            'RaidLevel': '50',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            }
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        target_raid_config = {
            'logical_disks':
                [
                    {
                        'size_gb': 100,
                        'raid_level': '50',
                    }
                ]
        }

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        elcm.create_raid_configuration(self.irmc_info,
                                       target_raid_config=target_raid_config)

        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_input data
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)

        raid_info_mock.assert_called_once_with(self.irmc_info)

    @mock.patch.object(elcm, 'elcm_profile_get')
    @mock.patch.object(elcm, '_create_raid_adapter_profile')
    def test_get_raid_config_with_logical_drives(
            self, create_raid_adapter_mock, elcm_profile_get_mock):

        profile_name = 'RAIDAdapter'
        elcm_profile_get_mock.return_value = {
            'Server': {
                'HWConfigurationIrmc': {
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                '@AdapterId': 'RAIDAdapter0',
                                '@ConfigurationType': 'Addressing',
                                'Arrays': {
                                    'Array': [
                                        {
                                            '@Number': 0,
                                            '@ConfigurationType': 'Addressing',
                                            'PhysicalDiskRefs': {
                                                'PhysicalDiskRef': [
                                                    {
                                                        '@Number': '1'
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'None',
                                            'RaidLevel': '0',
                                            'InitMode': 'fast',
                                            'Size': {
                                                '@Unit': 'GB',
                                                '#text': 100
                                            },
                                            'ArrayRefs': {
                                                'ArrayRef': [
                                                    {
                                                        '@Number': 0
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                                'PhysicalDisks': {
                                    'PhysicalDisk': [
                                        {
                                            '@Number': '0',
                                            '@Action': 'None',
                                            'Slot': 0,
                                            'PDStatus': 'Operational'
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }

        elcm.get_raid_adapter(irmc_info=self.irmc_info)
        elcm_profile_get_mock.assert_called_once_with(
            self.irmc_info, profile_name)
        create_raid_adapter_mock.assert_called_once_with(self.irmc_info)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, '_process_session_data')
    def test_delete_raid_adapter(
        self, session_mock, elcm_profile_delete_mock, elcm_profile_set_mock,
            raid_info_mock):
        raid_info_mock.return_value = {
            'Server': {
                'HWConfigurationIrmc': {
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                '@AdapterId': 'RAIDAdapter0',
                                '@ConfigurationType': 'Addressing',
                                'Arrays': {
                                    'Array': [
                                        {
                                            '@Number': 0,
                                            '@ConfigurationType': 'Addressing',
                                            'PhysicalDiskRefs': {
                                                'PhysicalDiskRef': [
                                                    {
                                                        '@Number': '1'
                                                    },
                                                    {
                                                        '@Number': '4'
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'None',
                                            'RaidLevel': '0',
                                            'ArrayRefs': {
                                                'ArrayRef': [
                                                    {
                                                        '@Number': 0
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                                'PhysicalDisks': {
                                    'PhysicalDisk': [
                                        {
                                            '@Number': '1',
                                            '@Action': 'None',
                                            'Slot': 1,
                                        },
                                        {
                                            '@Number': '4',
                                            '@Action': 'None',
                                            'Slot': 4,
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }

        expected_input = {
            'Server': {
                'HWConfigurationIrmc': {
                    '@Processing': 'execute',
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                '@AdapterId': 'RAIDAdapter0',
                                '@ConfigurationType': 'Addressing',
                                'Arrays': {
                                    'Array': [
                                        {
                                            '@Number': 0,
                                            '@ConfigurationType': 'Addressing',
                                            'PhysicalDiskRefs': {
                                                'PhysicalDiskRef': [
                                                    {
                                                        '@Number': '1'
                                                    },
                                                    {
                                                        '@Number': '4'
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                                'LogicalDrives': {
                                    'LogicalDrive': [
                                        {
                                            '@Number': 0,
                                            '@Action': 'Delete',
                                            'RaidLevel': '0',
                                            'ArrayRefs': {
                                                'ArrayRef': [
                                                    {
                                                        '@Number': 0
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                                'PhysicalDisks': {
                                    'PhysicalDisk': [
                                        {
                                            '@Number': '1',
                                            '@Action': 'None',
                                            'Slot': 1,
                                        },
                                        {
                                            '@Number': '4',
                                            '@Action': 'None',
                                            'Slot': 4,
                                        }
                                    ]
                                }
                            }
                        ]
                    },
                    '@Version': '1.00'
                },
                '@Version': '1.01'
            }
        }
        profile_name = 'RAIDAdapter'

        elcm_profile_set_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }
        session_id = 1
        session_timeout = 1800
        operation = 'CONFIG_RAID'

        elcm.delete_raid_configuration(irmc_info=self.irmc_info)
        session_mock.assert_called_once_with(self.irmc_info, operation,
                                             session_id, session_timeout)
        # Check raid_adapter data
        elcm_profile_set_mock.assert_called_once_with(
            self.irmc_info, expected_input)
        raid_info_mock.assert_called_once_with(self.irmc_info)
        elcm_profile_delete_mock.assert_called_once_with(self.irmc_info,
                                                         profile_name)

    @mock.patch.object(elcm, 'get_raid_adapter')
    @mock.patch.object(elcm, '_get_existing_logical_drives')
    @mock.patch.object(elcm, 'elcm_profile_set')
    @mock.patch.object(elcm, '_process_session_data')
    def test_delete_raid_adapter_without_existing_logical_drive(
            self, process_session_data_mock, elcm_profile_set_mock,
            existing_logical_drives_mock, raid_info_mock):
        raid_info_mock.return_value = {
            'Server': {
                'HWConfigurationIrmc': {
                    'Adapters': {
                        'RAIDAdapter': [
                            {
                                'key': 'value'
                            }
                        ]
                    }
                }
            }
        }

        existing_logical_drives_mock.return_value = None
        elcm.delete_raid_configuration(self.irmc_info)

        process_session_data_mock.assert_not_called()
        elcm_profile_set_mock.assert_not_called()

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_create')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    @mock.patch.object(elcm, 'elcm_session_delete')
    @mock.patch.object(elcm, 'elcm_profile_get')
    def test_success_session_monitoring(self, elcm_profile_get_mock,
                                        elcm_session_delete_mock,
                                        elcm_session_mock,
                                        elcm_profile_create_mock,
                                        elcm_profile_delete_mock):
        profile_name = 'RAIDAdapter'
        param_path = 'Server/HWConfigurationIrmc/Adapters/RAIDAdapter'
        session_id = 1

        elcm_profile_create_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }

        elcm_session_mock.return_value = {
            "Session": {
                "Id": 1,
                "Tag": "",
                "WorkSequence": "obtainProfileParameters",
                "Start": "2018\/03\/19 12:28:03",
                "Duration": 249,
                "Status": "terminated regularly"
            }
        }
        elcm._create_raid_adapter_profile(irmc_info=self.irmc_info)

        elcm_profile_get_mock.assert_called_once_with(self.irmc_info,
                                                      profile_name)
        elcm_session_delete_mock.assert_called_once_with(
            irmc_info=self.irmc_info, session_id=session_id, terminate=True)
        elcm_session_mock.assert_called_once_with(irmc_info=self.irmc_info,
                                                  session_id=session_id)
        elcm_profile_create_mock.assert_called_once_with(self.irmc_info,
                                                         param_path)
        elcm_profile_delete_mock.assert_called_once_with(self.irmc_info,
                                                         profile_name)

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_create')
    @mock.patch.object(elcm, 'elcm_session_get_status')
    @mock.patch.object(elcm, 'elcm_session_get_log')
    def test_fail_session_monitoring(self, elcm_session_get_log_mock,
                                     elcm_session_mock,
                                     elcm_profile_create_mock,
                                     elcm_profile_delete_mock):
        profile_name = 'RAIDAdapter'
        param_path = 'Server/HWConfigurationIrmc/Adapters/RAIDAdapter'

        session_id = 1

        elcm_profile_create_mock.return_value = {
            "Session": {
                "Id": 1,
                "A_Param": "abc123"
            }
        }

        elcm_session_mock.return_value = {
            "Session": {
                "Id": 1,
                "Tag": "",
                "WorkSequence": "obtainProfileParameters",
                "Start": "",
                "Duration": 0,
                "Status": "terminated - conflict with another running eLCM "
                          "activity"
            }
        }

        elcm_session_get_log_mock.return_value = {
            'SessionLog': {
                'Id': 1,
                'Tag': '',
                'WorkSequence': 'obtainProfileParameters',
                'Entries': {
                    'Entry': [
                        {
                            '@date': '2018\/03\/12 15:10:00',
                            '#text': 'createRaidDatabase: '
                                     'RAID Controller check start'
                        },
                        {
                            '@date': '2018\/03\/19 09:40:03',
                            '#text': 'LCMScheduler: Executing of '
                                     'obtainProfileParameters prohibited as '
                                     'obtainProfileParameters currently '
                                     'running'
                        }
                    ]
                }
            }
        }

        self.assertRaises(scci.SCCIClientError,
                          elcm._create_raid_adapter_profile, self.irmc_info)

        elcm_session_get_log_mock.assert_called_once_with(
            irmc_info=self.irmc_info, session_id=session_id)
        elcm_session_mock.assert_called_once_with(irmc_info=self.irmc_info,
                                                  session_id=session_id)
        elcm_profile_create_mock.assert_called_once_with(self.irmc_info,
                                                         param_path)
        elcm_profile_delete_mock.assert_called_once_with(self.irmc_info,
                                                         profile_name)

    @mock.patch.object(elcm, 'elcm_profile_delete')
    @mock.patch.object(elcm, 'elcm_profile_create')
    @mock.patch.object(elcm, '_process_session_data')
    def test_pass_raised_elcm_profile_not_found(
            self, _process_session_data_mock,
            elcm_profile_create_mock, elcm_profile_delete_mock):
        elcm_profile_delete_mock.side_effect = \
            elcm.ELCMProfileNotFound('not found')
        session_id = 1
        elcm_profile_create_mock.return_value = {
            'Session': {'Id': session_id,
                        'Status': 'running'}}
        session_timeout = 1800

        elcm._create_raid_adapter_profile(self.irmc_info)

        _process_session_data_mock.assert_called_once_with(
            self.irmc_info, 'CONFIG_RAID', session_id,
            session_timeout)
        elcm_profile_create_mock.assert_called_once_with(
            self.irmc_info, elcm.PARAM_PATH_RAID_CONFIG)
        elcm_profile_delete_mock.assert_called_once_with(
            self.irmc_info, elcm.PROFILE_RAID_CONFIG)

    @mock.patch.object(elcm, 'restore_bios_config')
    @mock.patch.object(elcm, 'elcm_profile_get_versions')
    def test_set_bios_configuration_without_versions(self,
                                                     get_versions_mock,
                                                     restore_bios_config_mock):
        settings = [{
            "name": "single_root_io_virtualization_support_enabled",
            "value": "True"
        }, {
            "name": "hyper_threading_enabled",
            "value": "True"
        }]

        get_versions_mock.return_value = {
            "Server": {
                "AdapterConfigIrmc": {
                    "@Version": "1.00"
                },
                "SystemConfig": {
                    "BiosConfig": {
                    }
                }
            }
        }

        bios_config_data = {
            'Server': {
                'SystemConfig': {
                    'BiosConfig': {
                        'PciConfig': {
                            'SingleRootIOVirtualizationSupportEnabled': True
                        },
                        'CpuConfig': {
                            'HyperThreadingEnabled': True,
                        }
                    }
                }
            }
        }
        elcm.set_bios_configuration(self.irmc_info, settings)
        restore_bios_config_mock.assert_called_once_with(self.irmc_info,
                                                         bios_config_data)

    @mock.patch.object(elcm, 'restore_bios_config')
    @mock.patch.object(elcm, 'elcm_profile_get_versions')
    def test_set_bios_configuration_with_versions(self,
                                                  get_versions_mock,
                                                  restore_bios_config_mock):
        settings = [{
            "name": "single_root_io_virtualization_support_enabled",
            "value": "True"
        }, {
            "name": "hyper_threading_enabled",
            "value": "True"
        }]

        get_versions_mock.return_value = {
            "Server": {
                "@Version": "1.01",
                "AdapterConfigIrmc": {
                    "@Version": "1.00"
                },
                "SystemConfig": {
                    "BiosConfig": {
                        "@Version": "1.02"
                    }
                }
            }
        }

        bios_config = {
            'Server': {
                'SystemConfig': {
                    'BiosConfig': {
                        'PciConfig': {
                            'SingleRootIOVirtualizationSupportEnabled': True
                        },
                        'CpuConfig': {
                            'HyperThreadingEnabled': True,
                        },
                        "@Version": "1.02"
                    }
                },
                "@Version": "1.01"
            }
        }
        elcm.set_bios_configuration(self.irmc_info, settings)
        restore_bios_config_mock.assert_called_once_with(self.irmc_info,
                                                         bios_config)

    @mock.patch.object(elcm, 'elcm_profile_get_versions')
    def test_set_bios_configuration_not_found(self,
                                              get_versions_mock):
        settings = [{
            "name": "single_root_io_virtualization_support_enabled",
            "value": "True"
        }, {
            "name": "setting1",
            "value": "True"
        }]

        get_versions_mock.return_value = {
            "Server": {
                "@Version": "1.01",
                "AdapterConfigIrmc": {
                    "@Version": "1.00"
                },
                "SystemConfig": {
                    "BiosConfig": {
                        "@Version": "1.02"
                    }
                }
            }
        }

        self.assertRaises(elcm.BiosConfigNotFound, elcm.set_bios_configuration,
                          self.irmc_info, settings)

    @mock.patch.object(elcm, 'restore_bios_config')
    @mock.patch.object(elcm, 'elcm_profile_get_versions')
    def test_set_bios_configuration_with_boolean_input(
            self, get_versions_mock, restore_bios_config_mock):
        settings = [{
            "name": "single_root_io_virtualization_support_enabled",
            "value": True
        }, {
            "name": "hyper_threading_enabled",
            "value": False
        }]

        get_versions_mock.return_value = {
            "Server": {
                "@Version": "1.01",
                "AdapterConfigIrmc": {
                    "@Version": "1.00"
                },
                "SystemConfig": {
                    "BiosConfig": {
                        "@Version": "1.02"
                    }
                }
            }
        }

        bios_config = {
            'Server': {
                'SystemConfig': {
                    'BiosConfig': {
                        'PciConfig': {
                            'SingleRootIOVirtualizationSupportEnabled': True
                        },
                        'CpuConfig': {
                            'HyperThreadingEnabled': False,
                        },
                        "@Version": "1.02"
                    }
                },
                "@Version": "1.01"
            }
        }
        elcm.set_bios_configuration(self.irmc_info, settings)
        restore_bios_config_mock.assert_called_once_with(self.irmc_info,
                                                         bios_config)

    @mock.patch.object(elcm, 'backup_bios_config')
    def test_get_bios_settings(self, backup_bios_config_mock):
        backup_bios_config_mock.return_value = {
            'bios_config': {
                "Server": {
                    "SystemConfig": {
                        "BiosConfig": {
                            "PciConfig": {
                                "SingleRootIOVirtualizationSupportEnabled":
                                    True,
                            },
                            "CpuConfig": {
                                "HyperThreadingEnabled": True,
                            }
                        }
                    }
                }
            }
        }
        result = elcm.get_bios_settings(self.irmc_info)
        expect_settings = [{
            "name": "single_root_io_virtualization_support_enabled",
            "value": "True"
        }, {
            "name": "hyper_threading_enabled",
            "value": "True"
        }]
        self.assertCountEqual(expect_settings, result)
        backup_bios_config_mock.assert_called_once_with(
            self.irmc_info)



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
SCCI functionalities shared between different iRMC modules.
"""

import functools
import re
import time

import defusedxml.ElementTree as ET
import requests

from scciclient.irmc import ipmi
from scciclient.irmc import snmp

DEBUG = False

S4_PATTERN = re.compile(r"^iRMC\s*S4$")
S5_PATTERN = re.compile(r"^iRMC\s*S5$")
FW_VERSION_HEAD_PATTERN = re.compile(r"^\d[\d.]*")
FW_VERSION_TAIL_PATTERN = re.compile(r"[a-zA-Z]*$")

S4_FD_SUPPORT_UPPER = 9.21
S5_FD_SUPPORT_UPPER = 1.25


class SCCIError(Exception):
    """SCCI Error

    This exception is general exception.
    """
    def __init__(self, message, errorcode=None):
        super(SCCIError, self).__init__(message)


class SCCIInvalidInputError(SCCIError):
    """SCCIInvalidInputError

    This exception is used when invalid inputs are passed to
    the APIs exposed by this module.
    """
    def __init__(self, message):
        super(SCCIInvalidInputError, self).__init__(message)


class SCCIClientError(SCCIError):
    """SCCIClientError

    This exception is used when a problem is encountered in
    executing an operation on the iRMC
    """
    def __init__(self, message):
        super(SCCIClientError, self).__init__(message)


class SCCIRAIDNotReady(SCCIError):
    """SCCIRAIDNotReady

    This exception is used when a mechanism not applied
    into a configuration on the iRMC yet
    """
    def __init__(self, message):
        super(SCCIRAIDNotReady, self).__init__(message)


class SCCISessionTimeout(SCCIError):
    def __init__(self, message):
        super(SCCISessionTimeout, self).__init__(message)


"""
List of iRMC S4/S5 supported SCCI commands

SCCI
OpCode  SCCI Command String      Description
0xE002  ConfigSpace              ConfigSpace Write
0x0111  PowerOnCabinet           Power On the Server
0x0112  PowerOffCabinet          Power Off the Server
0x0113  PowerOffOnCabinet        Power Cycle the Server
0x0204  ResetServer              Hard Reset the Server
0x020C  RaiseNMI                 Pulse the NMI (Non Maskable Interrupt)
0x0205  RequestShutdownAndOff    Graceful Shutdown, requires running Agent
0x0206  RequestShutdownAndReset  Graceful Reboot, requires running Agent
0x0209  ShutdownRequestCancelled Cancel a Shutdown Request
0x0203  ResetFirmware  Perform a BMC Reset
0x0251  ConnectRemoteFdImage     Connect or Disconnect a Floppy Disk image on a
                                 Remote Image Mount (NFS or CIFS Share )

                                 This command was deprecated at following
                                 version:
                                   iRMC S4: 9.62F
                                            (9.21F still supports virtual FD)
                                   iRMC S5: 1.60P
                                            (1.25P still supports virtual FD)

0x0252  ConnectRemoteCdImage     Connect or Disconnect a CD/DVD .iso image on a
                                 Remote Image Mount (NFS or CIFS Share )
0x0253  ConnectRemoteHdImage     Connect or Disconnect a Hard Disk image on a
                                 Remote Image Mount (NFS or CIFS Share )
"""

_POWER_CMD = '''
<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>
<CMDSEQ>
  <CMD Context="SCCI" OC="%s" OE="0" OI="0" Type="SET">
  </CMD>
</CMDSEQ>
'''


POWER_ON = _POWER_CMD % "PowerOnCabinet"
POWER_OFF = _POWER_CMD % "PowerOffCabinet"
POWER_CYCLE = _POWER_CMD % "PowerOffOnCabinet"
POWER_RESET = _POWER_CMD % "ResetServer"
POWER_RAISE_NMI = _POWER_CMD % "RaiseNMI"
POWER_SOFT_OFF = _POWER_CMD % "RequestShutdownAndOff"
POWER_SOFT_CYCLE = _POWER_CMD % "RequestShutdownAndReset"
POWER_CANCEL_SHUTDOWN = _POWER_CMD % "ShutdownRequestCancelled"


_VIRTUAL_MEDIA_CMD = '''
<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>
<CMDSEQ>
  <CMD Context="SCCI" OC="%s" OE="0" OI="0" Type="SET">
    <DATA Type="xsd::integer">%d</DATA>
  </CMD>
</CMDSEQ>
'''


MOUNT_CD = _VIRTUAL_MEDIA_CMD % ("ConnectRemoteCdImage", 1)
UNMOUNT_CD = _VIRTUAL_MEDIA_CMD % ("ConnectRemoteCdImage", 0)
MOUNT_FD = _VIRTUAL_MEDIA_CMD % ("ConnectRemoteFdImage", 1)
UNMOUNT_FD = _VIRTUAL_MEDIA_CMD % ("ConnectRemoteFdImage", 0)


_VIRTUAL_MEDIA_CD_SETTINGS = '''
<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>
<CMDSEQ>
  <!-- "ConfBmcMediaOptionsRemoteMediaEnabled" -->
  <!-- Make sure this one is enabled -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A80" OI="0" Type="SET">
    <DATA Type="xsd::integer">1</DATA>
  </CMD>
  <!-- "ConfBmcMediaOptionsCdNumber" -->
  <!-- Number of emulated CDROM/DVD Devices -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A68" OI="0" Type="SET">
    <DATA Type="xsd::integer">1</DATA>
  </CMD>
  <!-- "ConfBmcRemoteCdImageServer" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A60" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteCdImageUserDomain" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A63" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteCdImageShareType" -->
  <!-- 0 = NFS Share / 1 = CIFS Share -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A64" OI="0" Type="SET">
    <DATA Type="xsd::integer">%d</DATA>
  </CMD>
  <!-- "ConfBmcRemoteCdImageShareName" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A65" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteCdImageImageName" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A66" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteCdImageUserName" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A61" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteCdImageUserPassword" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A62" OI="0" Type="SET">
    <DATA Type="xsd::string" Encrypted="0">%s</DATA>
  </CMD>
</CMDSEQ>
'''


_VIRTUAL_MEDIA_FD_SETTINGS = '''
<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>
<CMDSEQ>
  <!-- "ConfBmcMediaOptionsRemoteMediaEnabled" -->
  <!-- Make sure this one is enabled -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A80" OI="0" Type="SET">
    <DATA Type="xsd::integer">1</DATA>
  </CMD>
  <!-- "ConfBmcMediaOptionsFdNumber" -->
  <!-- Number of emulated FD Devices -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A58" OI="0" Type="SET">
    <DATA Type="xsd::integer">1</DATA>
  </CMD>
  <!-- "ConfBmcRemoteFdImageServer" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A50" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteFdImageUserDomain" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A53" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteFdImageShareType" -->
  <!-- 0 = NFS Share / 1 = CIFS Share -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A54" OI="0" Type="SET">
    <DATA Type="xsd::integer">%d</DATA>
  </CMD>
  <!-- "ConfBmcRemoteFdImageShareName" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A55" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteFdImageImageName" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A56" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteFdImageUserName" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A51" OI="0" Type="SET">
    <DATA Type="xsd::string">%s</DATA>
  </CMD>
  <!-- "ConfBmcRemoteFdImageUserPassword" -->
  <CMD Context="SCCI" OC="ConfigSpace" OE="1A52" OI="0" Type="SET">
    <DATA Type="xsd::string" Encrypted="0">%s</DATA>
  </CMD>
</CMDSEQ>
'''


class MetaShareType(type):
    @property
    def nfs(cls):
        return cls.NFS

    @property
    def cifs(cls):
        return cls.CIFS


class ShareType(object, metaclass=MetaShareType):
    """"Virtual Media Share Type."""
    NFS = 0
    CIFS = 1


def get_share_type(share_type):
    """get share type."""
    return {
        'nfs': ShareType.nfs,
        'cifs': ShareType.cifs
    }[share_type.lower()]


def scci_cmd(host, userid, password, cmd, port=443, auth_method='basic',
             client_timeout=60, do_async=True, verify=False, **kwargs):
    """execute SCCI command

    This function calls SCCI server modules
    :param host: hostname or IP of iRMC
    :param userid: userid for iRMC with administrator privileges
    :param password: password for userid
    :param cmd: SCCI command
    :param port: port number of iRMC
    :param auth_method: irmc_username
    :param client_timeout: timeout for SCCI operations
    :param do_async: async call if True, sync call otherwise
    :param verify: (optional) Either a boolean, in which case it
                    controls whether we verify the server's TLS certificate,
                    or a string, in which case it must be a path to
                    a CA bundle to use. Defaults to ``False``.
    :returns: requests.Response from SCCI server
    :raises: SCCIInvalidInputError if port and/or auth_method params
             are invalid
    :raises: SCCIClientError if SCCI failed
    """
    auth_obj = None
    try:
        protocol = {80: 'http', 443: 'https'}[port]
        auth_obj = {
            'basic': requests.auth.HTTPBasicAuth(userid, password),
            'digest': requests.auth.HTTPDigestAuth(userid, password)
        }[auth_method.lower()]

    except KeyError:
        raise SCCIInvalidInputError(
            ("Invalid port %(port)d or " +
             "auth_method for method %(auth_method)s") %
            {'port': port, 'auth_method': auth_method})

    try:
        header = {'Content-type': 'application/x-www-form-urlencoded'}
        if kwargs.get('upgrade_type') == 'irmc':
            with open(cmd, 'rb') as file:
                data = file.read()
            config_type = '/irmcupdate?flashSelect=255'
        elif kwargs.get('upgrade_type') == 'bios':
            with open(cmd, 'rb') as file:
                data = file.read()
            config_type = '/biosupdate'
        else:
            # For EJECT command, validate parameters to handle abnormal case
            if check_eject_cd_cmd(cmd):
                if not validate_params_cd_fd("cmd_cd", protocol,
                                             host, auth_obj,
                                             do_async, client_timeout,
                                             verify):
                    return
            if check_eject_fd_cmd(cmd):
                if not validate_params_cd_fd("cmd_fd", protocol,
                                             host, auth_obj,
                                             do_async, client_timeout,
                                             verify):
                    return
            data = cmd
            config_type = '/config'

        r = requests.post(protocol + '://' + host + config_type,
                          data=data,
                          headers=header,
                          verify=verify,
                          timeout=client_timeout,
                          allow_redirects=False,
                          auth=auth_obj)

        if not do_async:
            time.sleep(5)
        else:
            # Async mode
            # Even in async mode, return immediately may cause error 5
            # (Input/Output error) for some commands such as POWER_ON,
            # POWER_OFF, POWER_RESET if we perform multiple calls of them
            # in a tight sequence or in parallel.
            # So, we sleep some time for such commands.
            if cmd in (POWER_ON, POWER_OFF, POWER_RESET):
                time.sleep(3)

        if DEBUG:
            print(cmd)
            print(r.text)
            print("do_async = %s" % do_async)
        if r.status_code not in (200, 201):
            raise SCCIClientError(
                ('HTTP PROTOCOL ERROR, STATUS CODE = %s' %
                 str(r.status_code)))

        result_xml = ET.fromstring(r.text)
        status = result_xml.find("./Value")
        # severity = result_xml.find("./Severity")
        error = result_xml.find("./Error")
        message = result_xml.find("./Message")
        if not int(status.text) == 0:
            raise SCCIClientError(
                ('SCCI PROTOCOL ERROR, STATUS CODE = %s, '
                 'ERROR = %s, MESSAGE = %s' %
                 (str(status.text), error.text, message.text)))
        else:
            return r

    except IOError as input_error:
        raise SCCIClientError(input_error)

    except ET.ParseError as parse_error:
        raise SCCIClientError(parse_error)

    except requests.exceptions.RequestException as requests_exception:
        raise SCCIClientError(requests_exception)


def validate_params_cd_fd(cmd_type, protocol, host, auth_obj,
                          do_async, client_timeout, verify):
    """Validate parameters of CD/DVD or FD Image Virtual Media in iRMC

    If one of parameters (ImageServer, ImageShareName or ImageName) set in
    ServerView Config Space is empty, and you try to eject virtual FD/CD,
    iRMC returns error. This function determines whether iRMC doesn't return
    error when you try to eject virtual FD/CD.
    :param cmd_type: command type has value switch between "cmd_cd" or "cmd_fd"
    :param protocol:
    :param host: hostname or IP of iRMC
    :param auth_obj: irmc userid/password
    :param do_async: async call if True, sync call otherwise
    :param client_timeout: timeout for SCCI operations
    :param verify: Either a boolean, in which case it
                   controls whether we verify the server's TLS certificate,
                   or a string, in which case it must be a path to
                   a CA bundle to use.
    :return: False if one of param is null. Otherwise, returns True.
    """

    if cmd_type == "cmd_cd":
        oe_image_server = "1A60"
        oe_image_server_share_name = "1A65"
        oe_image_name = "1A66"
    else:
        oe_image_server = "1A50"
        oe_image_server_share_name = "1A55"
        oe_image_name = "1A56"

    try:
        param = {'P45': '1', 'SAVE_DATA': '1'}
        header = {'Content-type': 'application/x-www-form-urlencoded'}
        r = requests.get(protocol + '://' + host + '/iRMC_Settings.pre',
                         params=param,
                         headers=header,
                         verify=verify,
                         timeout=client_timeout,
                         allow_redirects=False,
                         auth=auth_obj)
        if not do_async:
            time.sleep(5)

        if DEBUG:
            print("---------------------------")
            print("Current iRMC configuration:")
            print(r.url)

        if r.status_code not in (200, 201):
            raise SCCIClientError(
                ('HTTP PROTOCOL ERROR, STATUS CODE = %s' %
                 str(r.status_code)))

        result = r.text
        cmdseq = ET.fromstring(result)
        cfg_dict = {}
        for cmd_tag in cmdseq.iter(tag='CMD'):
            oe = cmd_tag.get('OE')
            data = cmd_tag.find('DATA').text
            cfg_dict[oe] = data

        if DEBUG:
            print("Server: ", cfg_dict[oe_image_server])
            print("Share Name: ", cfg_dict[oe_image_server_share_name])
            print("Image Name: ", cfg_dict[oe_image_name])
            print("---------------------------")

        if (cfg_dict[oe_image_server] is None) or \
                (cfg_dict[oe_image_server_share_name] is None) or \
                (cfg_dict[oe_image_name] is None):
            return False

    except ET.ParseError as parse_error:
        raise SCCIClientError(parse_error)
    except requests.exceptions.RequestException as requests_exception:
        raise SCCIClientError(requests_exception)

    return True


def check_eject_cd_cmd(xml_cmd):
    """To check command is MOUNT or UNMOUNT

    :param xml_cmd: the command
    :return: true if this is UNMOUNT command. Otherwise, return false.
    """

    try:
        cmdseq = ET.fromstring(xml_cmd.strip())
        cmd = cmdseq.find("./CMD")
        data = cmd.find("./DATA")
        if cmd.get("OC") == "ConnectRemoteCdImage" and \
                cmd.get("Type") == "SET" and data.text == "0":
            return True
    except ET.ParseError as parse_error:
        raise SCCIClientError(parse_error)
    return False


def check_eject_fd_cmd(xml_cmd):
    """To check command is MOUNT or UNMOUNT

    :param xml_cmd: the command
    :return: true if this is UNMOUNT command. Otherwise, return false.
    """
    try:
        cmdseq = ET.fromstring(xml_cmd.strip())
        cmd = cmdseq.find("./CMD")
        data = cmd.find("./DATA")
        if cmd.get("OC") == "ConnectRemoteFdImage" and \
                cmd.get("Type") == "SET" and data.text == "0":
            return True
    except ET.ParseError as parse_error:
        raise SCCIClientError(parse_error)
    return False


def get_client(host, userid, password, port=443, auth_method='basic',
               client_timeout=60, verify=False, **kwargs):
    """get SCCI command partial function

    This function returns SCCI command partial function
    :param host: hostname or IP of iRMC
    :param userid: userid for iRMC with administrator privileges
    :param password: password for userid
    :param port: port number of iRMC
    :param auth_method: irmc_username
    :param client_timeout: timeout for SCCI operations
    :param verify: (optional) Either a boolean, in which case it
                    controls whether we verify the server's TLS certificate,
                    or a string, in which case it must be a path to
                    a CA bundle to use. Defaults to ``False``.
    :returns: scci_cmd partial function which takes a SCCI command param
    """

    return functools.partial(scci_cmd, host, userid, password,
                             port=port, auth_method=auth_method,
                             client_timeout=client_timeout,
                             verify=verify, **kwargs)


def get_virtual_cd_set_params_cmd(remote_image_server,
                                  remote_image_user_domain,
                                  remote_image_share_type,
                                  remote_image_share_name,
                                  remote_image_deploy_iso,
                                  remote_image_username,
                                  remote_image_user_password):
    """get Virtual CD Media Set Parameters Command

    This function returns Virtual CD Media Set Parameters Command
    :param remote_image_server: remote image server name or IP
    :param remote_image_user_domain: domain name of remote image server
    :param remote_image_share_type: share type of ShareType
    :param remote_image_share_name: share name
    :param remote_image_deploy_iso: deploy ISO image file name
    :param remote_image_username: username of remote image server
    :param remote_image_user_password: password of the username
    :returns: SCCI command
    """

    cmd = _VIRTUAL_MEDIA_CD_SETTINGS % (
        remote_image_server,
        remote_image_user_domain,
        remote_image_share_type,
        remote_image_share_name,
        remote_image_deploy_iso,
        remote_image_username,
        remote_image_user_password)

    return cmd


def get_virtual_fd_set_params_cmd(remote_image_server,
                                  remote_image_user_domain,
                                  remote_image_share_type,
                                  remote_image_share_name,
                                  remote_image_floppy_fat,
                                  remote_image_username,
                                  remote_image_user_password):
    """get Virtual FD Media Set Parameters Command

    This function returns Virtual FD Media Set Parameters Command
    :param remote_image_server: remote image server name or IP
    :param remote_image_user_domain: domain name of remote image server
    :param remote_image_share_type: share type of ShareType
    :param remote_image_share_name: share name
    :param remote_image_deploy_iso: deploy ISO image file name
    :param remote_image_username: username of remote image server
    :param remote_image_user_password: password of the username
    :returns: SCCI command
    """
    cmd = _VIRTUAL_MEDIA_FD_SETTINGS % (
        remote_image_server,
        remote_image_user_domain,
        remote_image_share_type,
        remote_image_share_name,
        remote_image_floppy_fat,
        remote_image_username,
        remote_image_user_password)

    return cmd


def get_report(host, userid, password,
               port=443, auth_method='basic', client_timeout=60, verify=False):
    """get iRMC report

    This function returns iRMC report in XML format
    :param host: hostname or IP of iRMC
    :param userid: userid for iRMC with administrator privileges
    :param password: password for userid
    :param port: port number of iRMC
    :param auth_method: irmc_username
    :param client_timeout: timeout for SCCI operations
    :param verify: (optional) Either a boolean, in which case it
                    controls whether we verify the server's TLS certificate,
                    or a string, in which case it must be a path to
                    a CA bundle to use. Defaults to ``False``.
    :returns: root element of SCCI report
    :raises: ISCCIInvalidInputError if port and/or auth_method params
             are invalid
    :raises: SCCIClientError if SCCI failed
    """

    auth_obj = None
    try:
        protocol = {80: 'http', 443: 'https'}[port]
        auth_obj = {
            'basic': requests.auth.HTTPBasicAuth(userid, password),
            'digest': requests.auth.HTTPDigestAuth(userid, password)
        }[auth_method.lower()]

    except KeyError:
        raise SCCIInvalidInputError(
            ("Invalid port %(port)d or " +
             "auth_method for method %(auth_method)s") %
            {'port': port, 'auth_method': auth_method})

    try:
        r = requests.get(protocol + '://' + host + '/report.xml',
                         verify=verify,
                         timeout=(10, client_timeout),
                         allow_redirects=False,
                         auth=auth_obj)

        if r.status_code not in (200, 201):
            raise SCCIClientError(
                ('HTTP PROTOCOL ERROR, STATUS CODE = %s' %
                 str(r.status_code)))

        root = ET.fromstring(r.text)
        return root

    except ET.ParseError as parse_error:
        raise SCCIClientError(parse_error)

    except requests.exceptions.RequestException as requests_exception:
        raise SCCIClientError(requests_exception)


def get_sensor_data_records(report):
    """get sensor data

    This function returns sensor data in XML
    :param report: SCCI report element
    :returns: sensor element of SCCI report, or None
    """

    sensor = report.find("./System/SensorDataRecords")
    return sensor


def get_irmc_version(report):
    """get iRMC version

    This function returns iRMC version number
    :param report: SCCI report element
    :returns: version element of SCCI report, or None
    """

    version = report.find("./System/ManagementControllers/iRMC")
    return version


def get_irmc_version_str(report):
    """extract iRMC OS and iRMC firmware version from SCCI report

    This function returns iRMC OS name and iRMC firmware version
    :param report: SCCI report element
    :returns: a tuple of string (iRMC OS name, iRMC FW version)
    """

    version = get_irmc_version(report)
    return version.get('Name'), version.find('Firmware').text


def support_virtual_fd_str(irmc_os, fw_version):
    """determine iRMC supports virtual floppy disk

    This function determines whether iRMC supports virtual floppy disk
    based on provided iRMC OS & iRMC firmware version
    :param irmc_os: string representing iRMC OS
    :param fw_version: string representing iRMC firmware version
    :returns: boolean
    """

    version_head = FW_VERSION_HEAD_PATTERN.match(fw_version)

    if S4_PATTERN.match(irmc_os):
        if version_head and float(version_head.group()) <= S4_FD_SUPPORT_UPPER:
            return True
    elif S5_PATTERN.match(irmc_os):
        if version_head and float(version_head.group()) <= S5_FD_SUPPORT_UPPER:
            return True

    return False


def support_virtual_fd(report):
    """determine iRMC supports virtual floppy disk

    This function determines whether iRMC supports virtual floppy disk
    based on SCCI report
    :param report: SCCI report element
    :returns: boolean
    """

    irmc_os, fwv = get_irmc_version_str(report)
    return support_virtual_fd_str(irmc_os, fwv)


def get_essential_properties(report, prop_keys):
    """get essential properties

    This function returns a dictionary which contains keys as in
    prop_keys and its values from the report.

    :param report: SCCI report element
    :param prop_keys: a list of keys for essential properties
    :returns: a dictionary which contains keys as in
              prop_keys and its values.
    """
    v = {}
    v['memory_mb'] = int(report.find('./System/Memory/Installed').text)
    v['local_gb'] = sum(
        [int(int(size.text) / 1024)
         for size in report.findall('.//PhysicalDrive/ConfigurableSize')])
    v['cpus'] = sum([int(cpu.find('./CoreNumber').text)
                     for cpu in report.find('./System/Processor')
                     if cpu.find('./CoreNumber') is not None])
    # v['cpus'] = sum([int(cpu.find('./LogicalCpuNumber').text)
    #                 for cpu in report.find('./System/Processor')])
    v['cpu_arch'] = 'x86_64'

    return {k: v[k] for k in prop_keys}


def get_capabilities_properties(d_info,
                                capa_keys,
                                gpu_ids,
                                fpga_ids=None,
                                **kwargs):
    """get capabilities properties

    This function returns a dictionary which contains keys
    and their values from the report.


    :param d_info: the dictionary of ipmitool parameters for accessing a node.
    :param capa_keys: a list of keys for additional capabilities properties.
    :param gpu_ids: the list of string contains <vendorID>/<deviceID>
    for GPU.
    :param fpga_ids: the list of string contains <vendorID>/<deviceID>
    for CPU FPGA.
    :param kwargs: additional arguments passed to scciclient.
    :returns: a dictionary which contains keys and their values.
    """

    snmp_client = snmp.SNMPClient(
        address=d_info['irmc_address'],
        port=d_info['irmc_snmp_port'],
        version=d_info['irmc_snmp_version'],
        read_community=d_info['irmc_snmp_community'],
        user=d_info.get('irmc_snmp_user'),
        auth_proto=d_info.get('irmc_snmp_auth_proto'),
        auth_key=d_info.get('irmc_snmp_auth_password'),
        priv_proto=d_info.get('irmc_snmp_priv_proto'),
        priv_key=d_info.get('irmc_snmp_priv_password'))

    try:
        v = {}
        if 'rom_firmware_version' in capa_keys:
            v['rom_firmware_version'] = \
                snmp.get_bios_firmware_version(snmp_client)

        if 'irmc_firmware_version' in capa_keys:
            v['irmc_firmware_version'] = \
                snmp.get_irmc_firmware_version(snmp_client)

        if 'server_model' in capa_keys:
            v['server_model'] = snmp.get_server_model(snmp_client)

        # Sometime the server started but PCI device list building is
        # still in progress so system will response error. We have to wait
        # for some more seconds.
        if kwargs.get('sleep_flag', False) and \
                any(k in capa_keys for k in ('pci_gpu_devices', 'cpu_fpga')):
            time.sleep(5)

        if 'pci_gpu_devices' in capa_keys:
            v['pci_gpu_devices'] = ipmi.get_pci_device(d_info, gpu_ids)

        if fpga_ids is not None and 'cpu_fpga' in capa_keys:
            v['cpu_fpga'] = ipmi.get_pci_device(d_info, fpga_ids)

        if 'trusted_boot' in capa_keys:
            v['trusted_boot'] = ipmi.get_tpm_status(d_info)

        return v
    except (snmp.SNMPFailure, ipmi.IPMIFailure) as err:
        raise SCCIClientError('Capabilities inspection failed: %s' % err)


def process_session_status(irmc_info, session_timeout, upgrade_type):
    """process session for Bios config backup/restore or RAID config operation

    :param irmc_info: node info
    :param session_timeout: session timeout
    :param upgrade_type: flag to check upgrade with bios or irmc
    :return: a dict with following values:
        {
            'upgrade_message': <Message of firmware upgrade mechanism>,
            'upgrade_status'
        }
    """
    session_expiration = time.time() + session_timeout

    while time.time() < session_expiration:
        try:
            # Get session status to check
            session = get_firmware_upgrade_status(irmc_info, upgrade_type)
        except SCCIClientError:
            # Ignore checking during rebooted server
            time.sleep(10)
            continue

        status = session.find("./Value").text
        severity = session.find("./Severity").text
        message = session.find("./Message").text
        result = {}

        if severity == 'Information' and status != '0':
            if 'FLASH successful' in message:
                result['upgrade_status'] = 'Complete'
                return result
            # Sleep a bit
            time.sleep(5)
        elif severity == 'Error':
            result['upgrade_status'] = 'Error'
            return result
        else:
            # Error occurred, get session log to see what happened
            session_log = message
            raise SCCIClientError('Failed to set firmware upgrade. '
                                  'Session log is %s.' % session_log)

    else:
        raise SCCISessionTimeout('Failed to time out mechanism with %s.'
                                 % session_expiration)


def get_raid_fgi_status(report):
    """Gather fgi(foreground initialization) information of raid configuration

    This function returns a fgi status which contains activity status
    and its values from the report.

    :param report: SCCI report information
    :returns: dict of fgi status of logical_drives, such as Initializing (10%)
              or Idle. e.g: {'0': 'Idle', '1': 'Initializing (10%)'}
    :raises: SCCIInvalidInputError: fail report input.
             SCCIRAIDNotReady: waiting for RAID configuration to complete.
    """
    fgi_status = {}
    raid_path = "./Software/ServerView/ServerViewRaid"

    if not report.find(raid_path):
        raise SCCIInvalidInputError(
            "ServerView RAID not available in Bare metal Server")
    if not report.find(raid_path + "/amEMSV/System/Adapter/LogicalDrive"):
        raise SCCIRAIDNotReady(
            "RAID configuration not configure in Bare metal Server yet")

    logical_drives = report.findall(raid_path +
                                    "/amEMSV/System/Adapter/LogicalDrive")
    for logical_drive_name in logical_drives:
        status = logical_drive_name.find("./Activity").text
        name = logical_drive_name.find("./LogDriveNumber").text
        fgi_status.update({name: status})

    return fgi_status


def get_firmware_upgrade_status(irmc_info, upgrade_type):
    """get firmware upgrade status of bios or irmc

    :param irmc_info: dict of iRMC params to access the server node
        {
          'irmc_address': host,
          'irmc_username': user_id,
          'irmc_password': password,
          'irmc_port': 80 or 443, default is 443,
          'irmc_auth_method': 'basic' or 'digest', default is 'digest',
          'irmc_client_timeout': timeout, default is 60,
          'irmc_verify_ca': (optional) Either a boolean, in which case it
                            controls whether we verify the server's TLS
                            certificate, or a string, in which case it
                            must be a path to a CA bundle to use.
                            Defaults to ``False``.
          ...
        }
    :param upgrade_type: flag to check upgrade with bios or irmc
    :raises: ISCCIInvalidInputError if port and/or auth_method params
             are invalid
    :raises: SCCIClientError if SCCI failed
    """

    host = irmc_info.get('irmc_address')
    userid = irmc_info.get('irmc_username')
    password = irmc_info.get('irmc_password')
    port = irmc_info.get('irmc_port', 443)
    auth_method = irmc_info.get('irmc_auth_method', 'digest')
    client_timeout = irmc_info.get('irmc_client_timeout', 60)
    verify = irmc_info.get('irmc_verify_ca', False)

    auth_obj = None
    try:
        protocol = {80: 'http', 443: 'https'}[port]
        auth_obj = {
            'basic': requests.auth.HTTPBasicAuth(userid, password),
            'digest': requests.auth.HTTPDigestAuth(userid, password)
        }[auth_method.lower()]
    except KeyError:
        raise SCCIInvalidInputError(
            ("Invalid port %(port)d or " +
             "auth_method for method %(auth_method)s") %
            {'port': port, 'auth_method': auth_method})
    try:
        if upgrade_type == 'bios':
            upgrade_type = '/biosprogress'
        elif upgrade_type == 'irmc':
            upgrade_type = '/irmcprogress'
        r = requests.get(protocol + '://' + host + upgrade_type,
                         verify=verify,
                         timeout=(10, client_timeout),
                         allow_redirects=False,
                         auth=auth_obj)

        if r.status_code not in (200, 201):
            raise SCCIClientError(
                ('HTTP PROTOCOL ERROR, STATUS CODE = %s' %
                 str(r.status_code)))

        upgrade_status_xml = ET.fromstring(r.text)
        return upgrade_status_xml
    except ET.ParseError as parse_error:
        raise SCCIClientError(parse_error)
    except requests.RequestException as requests_exception:
        raise SCCIClientError(requests_exception)



# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import json
import time

from scciclient.irmc import elcm
from scciclient.irmc import scci


PROFILE_NAME = 'AdapterConfigIrmc'
PARAM_PATH = 'Server/AdapterConfigIrmc'


class ELCMVIOMClient(object):
    """Client calling eLCM REST APIs for VIOM feature"""

    def __init__(self, irmc_info):
        self.irmc_info = irmc_info

    def _wait_session(self, session_id, timeout=1800):
        session_expiration = time.time() + timeout

        while True:
            resp = elcm.elcm_session_get_status(self.irmc_info, session_id)

            status = resp['Session']['Status']
            if status == 'running' or status == 'activated':
                # Sleep a bit
                time.sleep(5)
            elif status == 'terminated regularly':
                return {}
            else:
                # Error occurred, get session log to see what happened
                try:
                    session_log = elcm.elcm_session_get_log(
                        irmc_info=self.irmc_info, session_id=session_id)
                except scci.SCCIClientError as e:
                    raise scci.SCCIClientError(
                        ('Operation Failed. Session %(session_id)s state is '
                         '%(session_state)s. Session log collection failed: '
                         '%(reason)s' %
                         {'session_id': session_id,
                          'session_state': resp['Session']['Status'],
                          'reason': e}))

                raise scci.SCCIClientError(
                    ('Operation failed. Session %(session_id)s state is '
                     '%(session_state)s. Session log is: "%(session_log)s".' %
                     {'session_id': session_id,
                      'session_state': resp['Session']['Status'],
                      'session_log': json.dumps(session_log)}))

            # Check for timeout
            if time.time() > session_expiration:
                # Timeout occurred, get session log to see what happened
                try:
                    session_log = elcm.elcm_session_get_log(
                        irmc_info=self.irmc_info, session_id=session_id)
                except scci.SCCIClientError as e:
                    raise elcm.ELCMSessionTimeout(
                        'Operation timed out. Session %(session_id)s has not '
                        'finished in %(timeout)d seconds. Session log '
                        'collection failed: %(reason)s' %
                        {'session_id': session_id,
                         'timeout': timeout,
                         'reason': e})

                raise elcm.ELCMSessionTimeout(
                    'Operation timed out. Session %(session_id)s has not '
                    'finished in %(timeout)d seconds. Session log is: '
                    '"%(session_log)s.' %
                    {'session_id': session_id,
                     'timeout': timeout,
                     'session_log': json.dumps(session_log)})

    def set_profile(self, adapter_config):
        _adapter_config = dict(adapter_config)
        _adapter_config.update({'@Processing': 'execute'})
        req = {'Server': {'AdapterConfigIrmc': _adapter_config,
                          '@Version': '1.01'}}
        resp = elcm.elcm_profile_set(self.irmc_info, req)
        self._wait_session(resp['Session']['Id'])

    def get_profile(self):

        # delete old one
        try:
            elcm.elcm_profile_delete(self.irmc_info, PROFILE_NAME)
        except elcm.ELCMProfileNotFound:
            pass

        resp = elcm.elcm_profile_create(self.irmc_info, PARAM_PATH)
        self._wait_session(resp['Session']['Id'])
        resp = elcm.elcm_profile_get(self.irmc_info, PROFILE_NAME)
        return resp


class VIOMAttribute(object):
    """Attribute in VIOM Element.

    This class is used for conversion between Python class and JSON table.
    """
    def __init__(self, name, key, init=None):
        self.name = name
        self.key = key
        self.init = init


class VIOMElement(object, metaclass=abc.ABCMeta):
    """Element in VIOM table."""
    def __init__(self, **kwargs):
        for attr in self.__class__._BASIC_ATTRIBUTES:
            setattr(self, attr.name, kwargs.get(attr.name, attr.init))

    def get_basic_json(self):
        table = {}
        for attr in self.__class__._BASIC_ATTRIBUTES:
            value = getattr(self, attr.name)
            if value is not None:
                table[attr.key] = value
        return table


class VIOMTable(VIOMElement):
    """Root class of VIOM table"""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('use_virtual_addresses', 'UseVirtualAddresses'),
        VIOMAttribute('viom_boot_enable', 'VIOMBootEnable'),
        VIOMAttribute('boot_menu_enable', 'BootMenuEnable'),
        VIOMAttribute('sriov', 'SRIOV'),
        VIOMAttribute('smux', 'Smux'),
        VIOMAttribute('boot_mode', 'BootMode'),
        VIOMAttribute('init_boot', 'InitBoot'),
        VIOMAttribute('processing', '@Processing'),
        VIOMAttribute('mode', 'Mode'),
        VIOMAttribute('version', '@Version', '1.00'),
    ]

    def __init__(self, **kwargs):
        super(VIOMTable, self).__init__(**kwargs)
        self.slots = {}
        self.manage = None

    def get_slot(self, slot_idx, create=True):
        slot = self.slots.get(slot_idx)
        if slot or not create:
            return slot
        slot = Slot(slot_idx)
        self.slots[slot_idx] = slot
        return slot

    def set_manage_table(self, manage):
        self.manage = manage

    def get_json(self):
        """Create JSON data for AdapterConfig.

        :returns: JSON data as follows:

            {
                "VIOMManage":{
                },
                "InitBoot":{
                },
                "UseVirtualAddresses":{
                },
                "BootMenuEnable":{
                },
                "SmuxSetting":{
                },
                "Slots":{
                }
            }
        """
        viom_table = self.get_basic_json()
        if self.slots:
            viom_table['Slots'] = {
                'Slot': [s.get_json() for s in self.slots.values()]
            }
        if self.manage:
            viom_table['VIOMManage'] = self.manage.get_json()
        return viom_table


class ManageTable(VIOMElement):
    """Class for ViomManage element."""

    _BASIC_ATTRIBUTES = [
        VIOMAttribute('manage', 'Manage'),
        VIOMAttribute('identification', 'Identification'),
        VIOMAttribute('trap_destination', 'TrapDestination'),
        VIOMAttribute('force', 'Force'),
        VIOMAttribute('preferred_version', 'PreferredInventoryVersion')
    ]

    def __init__(self, **kwargs):
        super(ManageTable, self).__init__(**kwargs)

    def get_json(self):
        """Create JSON data for ViomManage.

        :returns: JSON data for ViomManage as follows:

            {
                "Manage":{
                },
                "Force":{
                },
                "Identification":{
                },
                "TrapDestination":{
                },
                "PreferredInventoryVersion":{
                }
            }
        """
        return self.get_basic_json()


class Slot(VIOMElement):
    """Class for Slot element."""

    _BASIC_ATTRIBUTES = [
        VIOMAttribute('slot_idx', '@SlotIdx', 0),
    ]

    def __init__(self, slot_idx, **kwargs):
        super(Slot, self).__init__(slot_idx=slot_idx, **kwargs)
        self.onboard_cards = {}
        self.addon_cards = {}

    def add_card(self, card):
        if isinstance(card, OnboardCard):
            self.onboard_cards[card.card_idx] = card
        else:
            self.addon_cards[card.card_idx] = card

    def get_onboard_card(self, card_idx):
        return self.onboard_cards.get(card_idx)

    def get_addon_card(self, card_idx):
        return self.addon_cards.get(card_idx)

    def get_json(self):
        """Create JSON data for slot.

        :returns: JSON data for slot as follows:

            {
                "@SlotIdx":0,
                "OnboardControllers":{
                    "OnboardController": [
                    ]
                },
                "AddOnCards":{
                    "AddOnCard": [
                    ]
                }
            }
        """
        json = self.get_basic_json()
        if self.onboard_cards:
            json['OnboardControllers'] = {
                'OnboardController':
                    [c.get_json() for c in self.onboard_cards.values()]
            }
        if self.addon_cards:
            json['AddOnCards'] = {
                'AddOnCard': [c.get_json() for c in self.addon_cards.values()]
            }
        return json


class PCICard(object, metaclass=abc.ABCMeta):
    "Abstract class for PCI cards."
    def __init__(self, card_idx, adapter):
        self.card_idx = card_idx
        self.adapter = adapter

    def add_port(self, port):
        self.adapter.add_port(port)

    def get_port(self, port_idx):
        return self.adapter.get_port(port_idx)

    def get_json(self):
        """Create JSON data for PCI card element.

        :returns: JSON data for PCI card.
        Data for onboard card is as follows:

            {
                "@OnboardControllerIdx":1,
                "LANAdapter":{
                },
                "CNAAdapter":{
                }
            }

        Data for add-on card is as follows:

            {
                "@AddOnCardIdx":1,
                "LANAdapter":{
                },
                "FCAdapter":{
                },
                "CNAAdapter":{
                }
            }
        """
        json = {self.INDEX_KEY: self.card_idx}
        json.update(self.adapter.get_json())
        return json


class OnboardCard(PCICard):
    """Class for onboard Card."""
    INDEX_KEY = '@OnboardControllerIdx'


class AddOnCard(PCICard):
    """Class for add on card."""
    INDEX_KEY = '@AddOnCardIdx'


class Adapter(object, metaclass=abc.ABCMeta):
    """Abstract class for adapters.

    Adapter represents type of PCI card.
    """

    def __init__(self):
        self.ports = {}

    def add_port(self, port):
        self.ports[port.port_idx] = port

    def get_port(self, port_idx):
        return self.ports.get(port_idx)

    def get_json(self):
        """Create JSON data for adapter

        :returns: JSON data for adapter as follows:

            {
                "LANAdapter":{
                    "Ports":{
                        "Port": [
                        ]
                    }
                }
            }
        """
        return {
            self.ADAPTER_NAME: {
                'Ports': {
                    'Port': [p.get_json() for p in self.ports.values()]
                }
            }
        }


class LANAdapter(Adapter):
    """LAN adapter."""
    ADAPTER_NAME = 'LANAdapter'


class FCAdapter(Adapter):
    """FC adapter."""
    ADAPTER_NAME = 'FCAdapter'


class CNAAdapter(Adapter):
    """CNA adatper."""
    ADAPTER_NAME = 'CNAAdapter'


class AdapterPort(VIOMElement, metaclass=abc.ABCMeta):
    """Port in adapters."""

    def __init__(self, port_idx, **kwargs):
        super(AdapterPort, self).__init__(port_idx=port_idx, **kwargs)


class LANPort(AdapterPort):
    """LAN Port."""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('port_idx', '@PortIdx', 1),
        VIOMAttribute('port_enable', 'PortEnable'),
        VIOMAttribute('sriov', 'SRIOV'),
        VIOMAttribute('use_virtual_addresses', 'UseVirtualAddresses'),
    ]

    def __init__(self, port_idx, port_enable=True, mac=None, boot=None,
                 **kwargs):
        super(LANPort, self).__init__(port_idx, port_enable=port_enable,
                                      **kwargs)
        self.mac = mac
        self.boot = boot if boot else NoneBoot()

    def get_json(self):
        """Create JSON data for LANPort.

        :returns: JSON data as follows:

            {
              "@PortIdx":1,
              "PortEnable":{
              },
              "UseVirtualAddresses":{
              },
              "BootProtocol":{
              },
              "VirtualAddress":{
                "MAC":{
                }
              },
              "BootPriority":{
              },
              "ISCSIBootEnvironment":{
              }
            }
        """

        port = self.get_basic_json()
        port.update({
            'BootProtocol': self.boot.BOOT_PROTOCOL,
            'BootPriority': self.boot.boot_prio,
        })
        boot_env = self.boot.get_json()
        if boot_env:
            port.update(boot_env)
        if self.use_virtual_addresses and self.mac:
            port['VirtualAddress'] = {'MAC': self.mac}
        return port


class FCPort(AdapterPort):
    """FC Port."""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('port_idx', '@PortIdx', 1),
        VIOMAttribute('port_enable', 'PortEnable'),
        VIOMAttribute('sriov', 'SRIOV'),
        VIOMAttribute('use_virtual_addresses', 'UseVirtualAddresses'),
    ]

    def __init__(self, port_idx, port_enable=True, wwnn=None, wwpn=None,
                 boot=None, **kwargs):
        super(FCPort, self).__init__(port_idx, port_enable=port_enable,
                                     **kwargs)
        self.wwnn = wwnn
        self.wwpn = wwpn
        self.boot = boot if boot else NoneBoot()

    def get_json(self):
        """Create FC port.

        :returns: JSON for FC port as follows:
            {
                "@PortIdx":1,
                "PortEnable":{
                },
                "UseVirtualAddresses":{
                },
                "VirtualAddress":{
                    "WWNN":{
                    },
                    "WWPN":{
                    },
                    "MAC":{
                    }
                },
                "BootProtocol":{
                },
                "BootPriority":{
                },
                "FCBootEnvironment":{
                }
            }

        """
        port = self.get_basic_json()
        port.update({
            'BootProtocol': self.boot.BOOT_PROTOCOL,
            'BootPriority': self.boot.boot_prio,
        })
        boot_env = self.boot.get_json()
        if boot_env:
            port.update(boot_env)
        if self.use_virtual_addresses:
            addresses = {}
            if self.wwnn:
                addresses['WWNN'] = self.wwnn
            if self.wwpn:
                addresses['WWPN'] = self.wwpn
            if addresses:
                port['VirtualAddress'] = addresses
        return port


class CNAPort(AdapterPort):
    """CNA port."""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('port_idx', '@PortIdx', 1),
        VIOMAttribute('port_enable', 'PortEnable'),
    ]

    def __init__(self, port_idx, port_enable=True):
        super(CNAPort, self).__init__(port_idx, port_enable=port_enable)
        self.functions = {}

    def add_function(self, function):
        self.functions[function.func_idx] = function

    def get_function(self, func_idx):
        return self.functions.get(func_idx)

    def get_json(self):
        """Create JSON for CNA port.

        :returns: JSON for CNA port as follows:
            {
                "@PortIdx":1,
                "PortEnable":{
                },
                "Functions":{
                }
            }
        """
        port = self.get_basic_json()
        port['Functions'] = {
            'Function': [f.get_json() for f in self.functions.values()]
        }
        return port


class CNAFunction(VIOMElement, metaclass=abc.ABCMeta):
    """Abstract class for Functions for CNA card"""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('function_enable', 'FunctionEnable'),
        VIOMAttribute('vlan_id', 'VLANId'),
        VIOMAttribute('sriov', 'SRIOV'),
        VIOMAttribute('use_virtual_addresses', 'UseVirtualAddresses'),
        VIOMAttribute('bandwidth', 'Bandwidth'),
        VIOMAttribute('rate_limit', 'RateLimit'),
    ]

    def __init__(self, func_idx, function_enable=True, boot=None, **kwargs):
        super(CNAFunction, self).__init__(**kwargs)
        self.func_idx = func_idx
        self.boot = boot if boot else NoneBoot()
        self.function_enable = function_enable

    def _get_virtual_addresses_json(self, json):
        return None

    def get_json(self):
        """Create JSON for CNA function.

        :returns: JSON for CNA function.
            * LANFunction creates the following JSON:

                {
                    "LANFunction":{
                        "FunctionEnable":{
                        },
                        "BootProtocol":{
                        },
                        "UseVirtualAddresses":{
                        },
                        "BootPriority":{
                        },
                        "Bandwidth":{
                        },
                        "RateLimit":{
                        },
                        "VLANId":{
                        },
                        "VirtualAddress":{
                            "MAC":{
                            }
                        }
                    }
                }

            * FCoEFunction creates the following JSON:

                {
                    "FCoEFunction":{
                        "FunctionEnable":{
                        },
                        "BootProtocol":{
                        },
                        "UseVirtualAddresses":{
                        },
                        "BootPriority":{
                        },
                        "Bandwidth":{
                        },
                        "RateLimit":{
                        },
                        "VLANId":{
                        },
                        "VirtualAddress":{
                            "WWNN":{
                            },
                            "WWPN":{
                            },
                            "MAC":{
                            }
                        },
                        "FCBootEnvironment":{
                        }
                    }
                }

            * ISCSIFunction creates the following JSON:

                {
                    "@FunctionIdx": 1,
                    "ISCSIFunction":{
                        "FunctionEnable":{
                        },
                        "BootProtocol":{
                        },
                        "UseVirtualAddresses":{
                        },
                        "BootPriority":{
                        },
                        "Bandwidth":{
                        },
                        "RateLimit":{
                        },
                        "VLANId":{
                        },
                        "VirtualAddress":{
                            "MAC":{
                            }
                        },
                        "ISCSIBootEnvironment":{
                        }
                    }
                }
        """
        function = self.get_basic_json()
        function['BootProtocol'] = self.boot.BOOT_PROTOCOL
        function['BootPriority'] = self.boot.boot_prio
        if self.use_virtual_addresses:
            virtual_addresses = self._get_virtual_addresses_json()
            if virtual_addresses:
                function['VirtualAddress'] = virtual_addresses
        boot_env = self.boot.get_json()
        if boot_env:
            function.update(boot_env)
        return {'@FunctionIdx': self.func_idx,
                self.FUNCTION_NAME: function}


class LANFunction(CNAFunction):
    """LAN function for CNA card"""
    FUNCTION_NAME = 'LANFunction'

    def __init__(self, func_idx, function_enable=True, boot=None, mac=None,
                 **kwargs):
        super(LANFunction, self).__init__(
            func_idx, function_enable=function_enable, boot=boot, **kwargs)
        self.mac = mac

    def _get_virtual_addresses_json(self):
        return {'MAC': self.mac} if self.mac else None


class FCoEFunction(CNAFunction):
    """FCoE Function for CNA card."""
    FUNCTION_NAME = 'FCoEFunction'

    def __init__(self, func_idx, function_enable=True, boot=None, wwnn=None,
                 wwpn=None, mac=None, **kwargs):
        super(FCoEFunction, self).__init__(
            func_idx, function_enable=function_enable, boot=boot, **kwargs)
        self.wwnn = wwnn
        self.wwpn = wwpn
        self.mac = mac

    def _get_virtual_addresses_json(self):
        virtual_addresses = {}
        if self.mac:
            virtual_addresses['MAC'] = self.mac
        if self.wwnn:
            virtual_addresses['WWNN'] = self.wwnn
        if self.wwpn:
            virtual_addresses['WWPN'] = self.wwpn
        return virtual_addresses


class ISCSIFunction(CNAFunction):
    """iSCSI Function for CNA card."""
    FUNCTION_NAME = 'ISCSIFunction'

    def __init__(self, func_idx, function_enable=True, boot=None, mac=None,
                 **kwargs):
        super(ISCSIFunction, self).__init__(
            func_idx, function_enable=function_enable, boot=boot, **kwargs)
        self.mac = mac

    def _get_virtual_addresses_json(self):
        return {'MAC': self.mac} if self.mac else None


class Boot(VIOMElement, metaclass=abc.ABCMeta):
    """Abstract class for BootProtocol"""
    _BASIC_ATTRIBUTES = []

    def __init__(self, boot_prio=1, **kwargs):
        super(Boot, self).__init__(**kwargs)
        self.boot_prio = boot_prio

    def get_json(self):
        return {}


class NoneBoot(Boot):
    """None BootProtocol."""
    BOOT_PROTOCOL = 'None'


class PXEBoot(Boot):
    """PXE BootProtocol."""
    BOOT_PROTOCOL = 'PXE'


class FCBoot(Boot):
    """FC BootProtocol with FCBootEnvironment elemnt."""
    BOOT_PROTOCOL = 'FC'

    _BASIC_ATTRIBUTES = [
        VIOMAttribute('link_speed', 'FCLinkSpeed', 'auto'),
        VIOMAttribute('topology', 'FCTopology', 'auto_loop'),
        VIOMAttribute('boot_enable', 'SANBootEnable'),
    ]

    def __init__(self, boot_prio=1, **kwargs):
        super(FCBoot, self).__init__(boot_prio, **kwargs)
        self.targets = []

    def add_target(self, target):
        self.targets.append(target)

    def get_json(self):
        """Create JSON for FCBootEnvironment.

        :returns: JSON for FCBootEnvironment as follows:

            {
                "FCBootEnvironment":{
                    "FCTargets":{
                        "FCTarget":[
                        ]
                    },
                    "FCLinkSpeed":{
                    },
                    "SANBootEnable":{
                    },
                    "FCTopology":{
                    }
                }
            }
        """
        json = self.get_basic_json()
        for i in range(len(self.targets)):
            # @FCTargetIdx starts from 1.
            self.targets[i].set_index(i + 1)
        json['FCTargets'] = {
            'FCTarget': [t.get_json() for t in self.targets]
        }
        return {'FCBootEnvironment': json}


class FCTarget(VIOMElement):
    """FC Target."""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('index', '@FCTargetIdx', 1),
        VIOMAttribute('wwpn', 'TargetWWPN'),
        VIOMAttribute('lun', 'TargetLUN')
    ]

    def __init__(self, wwpn, lun=0, **kwargs):
        super(FCTarget, self).__init__(wwpn=wwpn, lun=lun)

    def set_index(self, index):
        self.index = index

    def get_json(self):
        """Create JSON for FCTarget.

        :returns: JSON data for FCTarget as follows:
            {
                "@FCTargetIdx":1,
                "TargetWWPN":{
                  },
                "TargetLUN":{
                }
            }
        """
        return self.get_basic_json()


class ISCSIBoot(Boot):
    """iSCSI BootProtocol with ISCSIBootEnvironment elment."""
    BOOT_PROTOCOL = 'ISCSI'

    def __init__(self, initiator, target, boot_prio=1):
        super(ISCSIBoot, self).__init__(boot_prio)
        self.initiator = initiator
        self.target = target

    def get_json(self):
        """Create JSON for ISCSIBoot.

        :returns: JSON data for ISCSIBoot as follows:
            {
                "ISCSIBootEnvironment":{
                    "ISCSIInitiator":{
                    },
                    "ISCSITarget":{
                    }
                }
            }
        """
        return {
            'ISCSIBootEnvironment': {
                'ISCSIInitiator': self.initiator.get_json(),
                'ISCSITarget': self.target.get_json()
            }
        }


class ISCSIInitiator(VIOMElement):
    """iSCSIInitiator."""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('dhcp_usage', 'DHCPUsage', False),
        VIOMAttribute('iqn', 'Name'),
        VIOMAttribute('ip', 'IPv4Address'),
        VIOMAttribute('subnet', 'SubnetMask'),
        VIOMAttribute('gateway', 'GatewayIPv4Address'),
        VIOMAttribute('vlan_id', 'VLANId', 0),
    ]

    def __init__(self, **kwargs):
        super(ISCSIInitiator, self).__init__(**kwargs)

    def get_json(self):
        """Create JSON data for iSCSI initiator.

        :returns: JSON data for iSCSI initiator as follows:

            {
                "DHCPUsage":{
                },
                "Name":{
                },
                "IPv4Address":{
                },
                "SubnetMask":{
                },
                "GatewayIPv4Address":{
                },
                "VLANId":{
                }
            }
        """
        if self.dhcp_usage:
            return {'DHCPUsage': self.dhcp_usage,
                    'Name': self.iqn}
        else:
            return self.get_basic_json()


class ISCSITarget(VIOMElement):
    """iSCSI target."""
    _BASIC_ATTRIBUTES = [
        VIOMAttribute('dhcp_usage', 'DHCPUsage', False),
        VIOMAttribute('iqn', 'Name'),
        VIOMAttribute('ip', 'IPv4Address'),
        VIOMAttribute('port', 'PortNumber', 3260),
        VIOMAttribute('lun', 'BootLUN', 0),
        VIOMAttribute('auth_method', 'AuthenticationMethod', 'None'),
        VIOMAttribute('chap_user', 'ChapUserName'),
        VIOMAttribute('chap_secret', 'ChapSecret'),
        VIOMAttribute('mutual_chap_secret', 'MutualChapSecret'),
    ]

    def __init__(self, **kwargs):
        super(ISCSITarget, self).__init__(**kwargs)

    def get_json(self):
        """Create JSON data for iSCSI target.

        :returns: JSON data for iSCSI target as follows:

            {
                "DHCPUsage":{
                },
                "Name":{
                },
                "IPv4Address":{
                },
                "PortNumber":{
                },
                "BootLUN":{
                },
                "AuthenticationMethod":{
                },
                "ChapUserName":{
                },
                "ChapSecret":{
                },
                "MutualChapSecret":{
                }
            }
        """
        json = {
            'DHCPUsage': self.dhcp_usage,
            'AuthenticationMethod': self.auth_method,
        }
        if not self.dhcp_usage:
            json['Name'] = self.iqn
            json['IPv4Address'] = self.ip
            json['PortNumber'] = self.port
            json['BootLUN'] = self.lun
        if self.chap_user:
            json['ChapUserName'] = self.chap_user
        if self.chap_secret:
            json['ChapSecret'] = self.chap_secret
        if self.mutual_chap_secret:
            json['MutualChapSecret'] = self.mutual_chap_secret
        return json
